/* See {float_image_align.h}. */
/* Last edited on 2008-05-25 03:22:40 by stolfi */

#define _GNU_SOURCE
#include <math.h>
#include <limits.h>
#include <assert.h>
#include <string.h>
 
#include <bool.h>
#include <r2.h>
#include <i2.h>
#include <jsmath.h>
#include <baltuple.h>
#include <affirm.h>
#include <float_image.h>
#include <float_image_align.h>

/* INTERNAL PROTOTYPES */

void fial_displace_points(int n, r2_t p[], i2_t d[], r2_t q[], r2_t step);
  /* Sets {q[0..n-1]} to the points {p[0..n-1]} displaced by the step counts
    {d[0..n-1]} times the given step lengths {step}. 

    More precisely sets {q[i].c[k] = p[i].c[k] + d[i].c[k]*step.c[k]}, 
    for {i} in {0..n-1} and {k} in {0,1}. */
    
void fial_debug_displacement(char *head, int n, i2_t d[], r2_t q[], double Q, char *tail);
  /* Prints the displacement step counts {d[0..n-1]}, the displaced
    points {q[0..n-1]}, and the mismatch value {Q} to {stderr},
    prefixed by {head} and siffixed by {tail}. */

/* IMPLEMENTATIONS */

void fial_optimal_shifts_brute
  ( int n,
    fial_mismatch_t *mm, /* Funtion that evaluates the mismatch between the images. */
    r2_t step,           /* Precision required along each axis. */
    i2_t ns,             /* Max number of adjustment steps along each axis. */
    r2_t p[],            /* (IN/OUT) Corresponding points in each image. */
    double *QP           /* (OUT) Discrepancy measure for best alignment. */
  )
  { 
    bool_t debug = TRUE;
    int i, k;
      
    /* Check search grid parameters, set {smin[0..1],smax[0..1]}: */
    int smin[2]; /* Min sum of displacements in each coord. */
    int smax[2]; /* Max sum of displacements in each coord. */
    for (k = 0; k < 2; k++)
      { demand(ns.c[k] >= 0, "search step count must be non-negative");
        if (ns.c[k] > 0)
          { demand(step.c[k] > 0, "search {step} must be positive"); }
        smin[k] = -ns.c[k]/2;
        smax[k] = smin[k] + ns.c[k] - 1;
      }
    
    i2_t dbest[n];       /* Best Adjustment step counts found so far. */
    double Qbest = INF;  /* Mismatch of best alignment. */
        
    /* Enumeration loop: */
    int dx[n];          /* Adjustment step counts in X direction. */
    fial_first_tuple(n, dx, ns.c[0], smin[0], smax[0]);
    do
      { int dy[n];          /* Adjustment step counts in Y direction. */
        fial_first_tuple(n, dy, ns.c[1], smin[1], smax[1]);
        do
          { /* Merge {dx,dy} into {d[0..n-1]}: */
            i2_t d[n];
            for (i = 0; i < n; i++)  { d[i] = (i2_t){{ dx[i], dy[i] }}; }
            /* Evaluate {Q} for the adjustment step count tuple {d[0..n-1]}: */
            r2_t q[n];      /* Work: Displaced points {q[i] = p[i] + d[i]}. */
            fial_displace_points(n, p, d, q, step); 
            double Q = mm(n, q);
            if (debug) { fial_debug_displacement("trial", n, d, q, Q, "\n"); }
            if (Q < Qbest) 
              { Qbest = Q; for (i = 0; i < n; i++)  { dbest[i] = d[i]; } } 
          }
        while (! fial_next_tuple(n, dy, ns.c[0], smin[1], smax[1]));
      }
    while (! fial_next_tuple(n, dx, ns.c[0], smin[0], smax[0]));
    /* Enumeration finished, return the best solution: */
    fial_displace_points(n, p, dbest, p, step); 
    (*QP) = Qbest;
  }    

void fial_displace_points(int n, r2_t p[], i2_t d[], r2_t q[], r2_t step)
  { int i, k;
    for (i = 0; i < n; i++)
      { for (k = 0; k < 2; k++)
          { q[i].c[k] = p[i].c[k] + d[i].c[k]*step.c[k]; }
      }
  }

void fial_debug_displacement(char *head, int n, i2_t d[], r2_t q[], double Q, char *tail)
  {
    fprintf(stderr, "    %s\n", head);
    int i;
    i2_t s = (i2_t){{0,0}};
    for (i = 0; i < n; i++)
      { if (d != NULL) 
          { fprintf(stderr, "      d[%02d] = ( %+3d  %+3d )", i, d[i].c[0], d[i].c[1]); 
            s.c[0] += d[i].c[0];
            s.c[1] += d[i].c[1];
          }
        if (q != NULL) 
          { fprintf(stderr, "  q[%02d] = ( %+9.2f  %+9.2f )", i, q[i].c[0], q[i].c[1]); }
        fprintf(stderr, "\n");
      }
    if (d != NULL) 
      { fprintf(stderr, "      sum =   ( %+3d  %+3d )", s.c[0], s.c[1]); }
    fprintf(stderr, "    Q = %12.5f\n", Q);
    fprintf(stderr, "%s", tail);
  }

