/* See {float_image_mscale.h}. */
/* Last edited on 2009-02-24 00:21:24 by stolfi */

#define _GNU_SOURCE
#include <assert.h>
#include <limits.h>
#include <string.h>
#include <math.h>
 
#include <bool.h>
#include <affirm.h>
#include <r2.h>
#include <wt_table.h>
#include <float_image.h>

#include <float_image_mscale.h>

/* INTERNAL PROTOTYPES */

/* IMPLEMENTATIONS */

float_image_t *float_image_mscale_shrink(float_image_t *A, float_image_t *M, int nw)
  { 
    /* Generate the 1D weight mask: */
    demand(nw > 0, "invalid filter width");
    double wt[nw];
    wt_table_fill_binomial(nw, wt);
    
    /* Get the image dimentsions: */
    int NC = A->sz[0];
    int NXA = A->sz[1];
    int NYA = A->sz[2];

    /* Check the mask dimentsions: */
    if (M != NULL)
      { demand(M->sz[0] == 1, "weight mask should be monochromatic");
        demand(M->sz[1] == NXA, "mask width mismatch");
        demand(M->sz[2] == NYA, "mask height mismatch");
      }
    
    /* Compute the output image size: */
    int NXR = (NXA+1)/2;
    int NYR = (NYA+1)/2;
    
    /* Compute the window displacement {hw}: */
    int hw = (nw-1)/2;
    
    /* Create the output image {R}: */
    float_image_t *R = float_image_new(NC, NXR, NYR);
    
    /* Fill the pixels of {R}: */
    int xR, yR, c;
    for(yR = 0; yR < NYR; yR++)
      { for(xR = 0; xR < NXR; xR++)
          { /* Accumulate the weighted pixel sum over the window: */
            double sum_w = 0;    /* Sum of weights. */
            double sum_w_v[NC];  /* Sum of weighted pixels. */
            for (c = 0; c < NC; c++) { sum_w_v[c] = 0; }
            int xD, yD;
            for(yD = 0; yD < nw; yD++)
              { int yA = 2*yR+yD-hw;
                double wty = ((yA < 0) || (yA >= NYA) ? 0.0 : wt[yD]);
                for(xD = 0; xD < nw; xD++)
                  { int xA = 2*xR+xD-hw;
                    double wtx = ((xA < 0) || (xA >= NXA) ? 0.0 : wt[xD]);
                    double w = wtx*wty;
                    if (w > 0)
                      { /* Multiply by the mask weight, if any: */
                        if (M != NULL) { w *= float_image_get_sample(M, 0, xA, yA); }
                        sum_w += w; 
                        for (c = 0; c < NC; c++)
                          { double v = float_image_get_sample(A, c, xA, yA);
                            sum_w_v[c] += w*v;
                          }
                      }
                  }
              }
            /* Store the weighted average (maybe NAN) in the {R} pixel: */
            for (c = 0; c < NC; c++)
              { double v = sum_w_v[c]/sum_w; 
                float_image_set_sample(R, c, xR, yR, v);
              }
          }
      }
    return R;
  }
  
float_image_t *float_image_mscale_mask_shrink(float_image_t *M, int nw, bool_t harm)
  { /* Generate the 1D weight mask: */
    demand(nw > 0, "invalid filter width");
    double wt[nw];
    wt_table_fill_binomial(nw, wt);
    
    /* Get the image dimentsions: */
    int NC = M->sz[0];
    int NXM = M->sz[1];
    int NYM = M->sz[2];
    
    /* Compute the output image size: */
    int NXR = (NXM+1)/2;
    int NYR = (NYM+1)/2;
    
    /* Create the output image {R}: */
    float_image_t *R = float_image_new(NC, NXR, NYR);
    
    /* Compute the window displacement {hw}: */
    int hw = (nw-1)/2;
    
    /* Fill the pixels of {R}: */
    int xR, yR, c;
    for(yR = 0; yR < NYR; yR++)
      { for(xR = 0; xR < NXR; xR++)
          { for (c = 0; c < NC; c++)
              { /* Accumulate the sample sums over the window: */
                double sum_w = 0;    /* Sum of weights. */
                double sum_w_v = 0;  /* Sum of weighted pixels. */
                int xD, yD;
                for(yD = 0; yD < nw; yD++)
                  { int yM = 2*yR+yD-hw;
                    double wty = ((yM < 0) || (yM >= NYM) ? 0.0 : wt[yD]);
                    for(xD = 0; xD < nw; xD++)
                      { int xM = 2*xR+xD-hw;
                        double wtx = ((xM < 0) || (xM >= NXM) ? 0.0 : wt[xD]);
                        double w = wtx*wty;
                        if (w > 0)
                          { sum_w += w;
                            double v = float_image_get_sample(M, c, xM, yM);
                            /* !!! Must find a probabilistic justification for this: */
                            sum_w_v += (harm ? w/v : w*v);
                          }
                      }
                  }
                double v = (harm ? sum_w/sum_w_v : sum_w_v/sum_w);
                float_image_set_sample(R, c, xR, yR, v);
              }
          }
      }
    return R;
  }

r2_t float_image_mscale_point_shrink(r2_t *pA, int nw)
  {
    if ((nw & 1) != 0)
      { /* Odd: */ return (r2_t){{ 0.5*(pA->c[0]+0.5), 0.5*(pA->c[1]+0.5) }}; }
    else
      { /* Evn: */ return (r2_t){{ 0.5*pA->c[0], 0.5*pA->c[1] }}; }
  }

r2_t float_image_mscale_point_expand(r2_t *pR, int nw)
  {
    if ((nw & 1) != 0)
      { /* Odd: */ return (r2_t){{ 2.0*pR->c[0] - 0.5, 2.0*pR->c[1] - 0.5 }}; }
    else
      { /* Evn: */ return (r2_t){{ 2.0*pR->c[0], 2.0*pR->c[1] }}; }
  }

int float_image_mscale_rounding_bias(int n)
{
  int b = 0;
  while (n & 1) { b = 1-b; n /= 2; }
  return b;
}
 
