#define PROG_NAME "test_indexing"
#define PROG_DESC "test of {indexing.h}"
#define PROG_VERS "1.0"

/* Last edited on 2009-03-06 10:06:56 by stolfi */ 
/* Created on 2005-02-14 (or earlier) by J. Stolfi, UNICAMP */

#define test_indexing_COPYRIGHT \
  "Copyright  2005  by the State University of Campinas (UNICAMP)"

#include <stdio.h>
#include <stdlib.h>
#include <stdint.h>
#include <string.h>
#include <assert.h>

#include <indexing.h>
#include <affirm.h>
#include <jsrandom.h>

#define bug(FMT_AND_ARGS...) \
  do { \
    fprintf(stderr, "%s:%d: ", __FILE__, __LINE__); \
    fprintf(stderr, FMT_AND_ARGS); \
    fprintf(stderr, "\n"); \
    exit(1); \
  } while(0)

#define MAXDIM 6

typedef struct desc_t
  { ix_dim_t d;
    ix_size_t z[MAXDIM];
    ix_pos_t b;
    ix_step_t s[MAXDIM];
    int nel;     /* Number of elements allocated for  {el}. */
    uint8_t *el; /* Element storage area. */
  } desc_t;

desc_t *make_desc(ix_dim_t d, ix_size_t sz[]);
  /* Creates a descriptor with axis count {d} and
    size vector {sz[0..d-1]}.  Some indices {i}
    may have virtual replication ({st[i] == 0}
    with {sz[i] > 1}). */

desc_t *make_packed_desc(ix_dim_t d, ix_size_t sz[], ix_order_t ixor);
  /* Creates a descriptor with axis count {d} and
    size vector {sz[0..d-1]}, with steps consistent
    with a packed element arrangement with index
    order {L}, and no virtual replication. */

void do_test_suite(void);

void test_size_ops(void);
/* void test_new_array(desc_t *A, ix_size_t *sz); */
void test_addr(desc_t *A);
void test_enum(desc_t *A);
void test_packed(desc_t *A);
/* void test_indices(desc_t *A); */
void test_crop(desc_t *A);
void test_subsample(desc_t *A);
void test_swap_indices(desc_t *A);
void test_flip_indices(desc_t *A);
void test_flip(desc_t *A);
void test_diagonal(desc_t *A);
void test_chop(desc_t *A);
void test_replicate(desc_t *A);
void print_desc(FILE *wr, char *pf, desc_t *A, char *sf);

/* void dump_storage(FILE *wr, void *el, int nw, ix_nbits_t bpw); */
void check_sizes(ix_dim_t d, ix_size_t *sza, ix_size_t *szb);
void check_max_indices(ix_dim_t d, ix_index_t *ma, ix_index_t *mb);
void check_max_indices_sz(ix_dim_t d, ix_index_t *ma, ix_size_t *szb);

int main (int argn, char **argv)
{
  
  /* General limit and data type tests: */
  demand((ix_pos_NONE < 0) || (ix_pos_NONE > ix_MAX_POS), "invalid ix_pos_NONE");

  do_test_suite();
  return 0;
}
 
void do_test_suite(void)
{
  test_size_ops();
  
  /* An array for testing: */
  ix_dim_t d = 5;
  ix_size_t sz[MAXDIM];
  sz[0] = 3; sz[1] = 10; sz[2] = 1; sz[3] = 30; sz[4] = 15;
  desc_t *A = make_desc(d, sz);
  
  /* test_new_array(A, sz, bps, bpw); */
  test_packed(A);
  test_enum(A);
  test_addr(A);
  test_crop(A);
  test_subsample(A);
  test_swap_indices(A);
  test_flip_indices(A);
  test_flip(A);
  test_diagonal(A);
  test_chop(A);
  test_replicate(A);
}

desc_t *make_desc(ix_dim_t d, ix_size_t sz[])
{
  desc_t *A = notnull(malloc(sizeof(desc_t)), "no mem");
  ix_pos_t base = 1000*abrandom(1,9);
  ix_axis_t iz = abrandom(1,d-2);     /* Index with zero step. */
  A->d = d;
  int i;
  int64_t npos = 1;
  for (i = 0; i < d; i++)
    { A->z[i] = sz[i];
      if (i == iz)
        { /* Use virtual replication on this index: */
          A->s[i] = 0;
        }
      else
        { /* Use distinct elements on this index: */
          A->s[i] = (sz[i] == 1 ? 0 : npos); 
          npos *= sz[i];
        }
    }
  A->b = base;
  return A;
}

desc_t *make_packed_desc(ix_dim_t d, ix_size_t sz[], ix_order_t ixor)
{
  desc_t *A = notnull(malloc(sizeof(desc_t)), "no mem");
  A->d = d;
  ix_sizes_assign(d, A->z, sz);
  ix_packed_steps(d, sz, ixor, A->s);
  A->b = 0;
  return A;
}

void test_size_ops(void)
{
  fprintf(stderr, "Checking ix_sizes_assign, ix_sizes_shrink, ix_sizes_expand...\n");
  ix_dim_t d = 4; affirm(d < MAXDIM, "bad d or MAXDIM"); /* A dim less than {MAXDIM}. */
  bool_t ch_r, ch_g; 

  /* Size vectors for tests: */
  ix_size_t sza[MAXDIM];
  ix_size_t szb[MAXDIM];
  ix_size_t szc[MAXDIM];
  ix_size_t szd[MAXDIM];
  
  /* Initialize {sza,szb,szc} with random values: */
  auto ix_size_t rnd_sz(int a, int b, int c, int i); 
    /* A funct of {a,b,c}, and {i}. */

  ix_size_t rnd_sz(int a, int b, int c, int i) 
    { return a + ((b*i*i) % c); }
  
  int i;
  for (i = 0; i < MAXDIM; i++)
    { sza[i] = rnd_sz(17,  8, 13, i);
      szb[i] = rnd_sz(19, 10, 17, i);
      szc[i] = rnd_sz(23, 12, 19, i);
      szd[i] = rnd_sz(29, 16, 23, i);
    }

  /* Test {ix_size_assign}: */
  ix_sizes_assign(d, sza, szb);
  for (i = 0; i < MAXDIM; i++)
    { ix_size_t g = (i < d ? szb[i] : rnd_sz(17,  8, 13, i)); 
      if (sza[i] != g) 
        { bug("ix_size_assign mismatch i = %d sza = %llu should be %llu", i, sza[i], g); }
    }
    
  /* Test {ix_size_min}: */
  ix_sizes_assign(MAXDIM, szb, sza); /* Save current value of {sza} in {szb}. */
  ch_r = ix_sizes_shrink(d, sza, szc);
  ch_g = FALSE;
  for (i = 0; i < MAXDIM; i++)
    { ix_size_t g = (i < d ? (szb[i] < szc[i] ? szb[i] : szc[i]) : szb[i]); 
      if (sza[i] != g) 
        { bug("ix_size_min mismatch i = %d sza = %llu should be %llu", i, sza[i], g); }
      if ((i < d) && (sza[i] != szb[i])) { ch_g = TRUE; }
    }
  if (ch_r != ch_g)
    { bug("ix_size_min wrong return value"); }
    
  /* Test {ix_size_max}: */
  ix_sizes_assign(MAXDIM, szb, sza); /* Save current value of {sza} in {szb}. */
  ch_r = ix_sizes_expand(d, sza, szd);
  ch_g = FALSE;
  for (i = 0; i < MAXDIM; i++)
    { ix_size_t g = (i < d ? (szb[i] > szd[i] ? szb[i] : szd[i]) : szb[i]); 
      if (sza[i] != g) 
        { bug("ix_size_max mismatch i = %d sza = %llu should be %llu", i, sza[i], g); }
      if ((i < d) && (sza[i] != szb[i])) { ch_g = TRUE; }
    }
  if (ch_r != ch_g)
    { bug("ix_size_max wrong return value"); }
}

void test_addr(desc_t *A)
{
  fprintf(stderr, "Checking ix_sample_addr, setting and checking all elems...\n");
  /* Clear elements: */
  int i;
  for (i = 0; i < A->nel; i++) { A->el[i] = 0; }
  bool_t verbose = FALSE;
  if (verbose) { print_desc(stderr, "A = { ", A, " }\n"); } 
  ix_pos_t addr_exp = A->b;
  ix_index_t ixA[MAXDIM];

  /* Allocate a large enough boolean array: */
  int64_t npos = 1;
  { int i; for (i = 0; i < A->d; i++) { npos *= A->z[i]; } }
  int64_t nel = A->b + npos + 1000;
  uint8_t *el = (uint8_t *)notnull(malloc(nel*sizeof(uint8_t)), "no mem");

  for (ixA[4] = 0; ixA[4] < A->z[4]; ixA[4]++)
    for (ixA[3] = 0; ixA[3] < A->z[3]; ixA[3]++)
      for (ixA[2] = 0; ixA[2] < A->z[2]; ixA[2]++)
        for (ixA[1] = 0; ixA[1] < A->z[1]; ixA[1]++)
          for (ixA[0] = 0; ixA[0] < A->z[0]; ixA[0]++)
            { 
              if (! ix_is_valid(A->d, ixA, A->z)) { bug("not ix_is_valid"); } 
              ix_pos_t addr = ix_position(A->d, ixA, A->b, A->s);
              if (addr != addr_exp) 
                { bug("ix_position mismatch addr = %llu addr_exp = %llu", addr, addr_exp); } 
              ix_pos_t addr_safe = ix_position_safe(A->d, ixA, A->z, A->b, A->s);
              if (addr != addr_safe) 
                { bug("ix_position_safe mismatch addr = %llu addr_safe = %llu", addr, addr_safe); } 
              uint8_t smp_exp = 1 + (addr % 253);
              el[addr] = smp_exp;
              addr_exp++;
            }

  fprintf(stderr, "Re-checking the value of all samples...\n");
  for (ixA[4] = 0; ixA[4] < A->z[4]; ixA[4]++)
    for (ixA[3] = 0; ixA[3] < A->z[3]; ixA[3]++)
      for (ixA[2] = 0; ixA[2] < A->z[2]; ixA[2]++)
        for (ixA[1] = 0; ixA[1] < A->z[1]; ixA[1]++)
          for (ixA[0] = 0; ixA[0] < A->z[0]; ixA[0]++)
            { 
              ix_pos_t addr = ix_position(A->d, ixA, A->b, A->s);
              uint8_t smp_exp = 1 + (addr % 253);
              uint8_t smp = el[addr];
              if (smp != smp_exp) 
                { bug("set/get sample mismatch smp = %u smp_exp = %u", smp, smp_exp); } 
            }
}

void test_enum(desc_t *A)
{
  fprintf(stderr, "Checking ix_enum...\n");
  
  /* Fabricate the other two arrays: */
  ix_dim_t d = A->d;
  ix_size_t *sz = A->z;
  desc_t *B = make_desc(d, sz);
  desc_t *C = make_desc(d, sz);
  bool_t verbose = FALSE;
  if (verbose) 
    { print_desc(stderr, "A = { ", A, " }\n");
      print_desc(stderr, "B = { ", B, " }\n");
      print_desc(stderr, "C = { ", C, " }\n");
    }
    
  ix_index_t ix[MAXDIM];

  /* Allocate a vector to store all positions of {A}: */
  ix_pos_count_t npos = ix_num_tuples(d, sz);
  ix_pos_t *posA = notnull(malloc(npos*sizeof(ix_pos_t)), "no mem");
  ix_pos_t *posB = notnull(malloc(npos*sizeof(ix_pos_t)), "no mem");
  ix_pos_t *posC = notnull(malloc(npos*sizeof(ix_pos_t)), "no mem");
  int64_t kpos;     /* Current index into {posA}. */

  auto void check_fw( ix_index_t ix[], ix_pos_t pA, ix_pos_t pB, ix_pos_t pC );
  auto void check_bw( ix_index_t ix[], ix_pos_t pA, ix_pos_t pB, ix_pos_t pC );
    /* These are procedures that can be passed as the {op} argument of 
      {ix_enum}. They check whether the given positions {pA,pB,pC}
      match {posA[kpos],posB[kpos],posC[kpos]}.  The procedure 
      {check_fw} increments {kpos} after the test,
      while {check_bw} decrements it before the test. */

  /* Enumerate all positions by hand in "F" order: */
  kpos = 0;
  for (ix[4] = 0; ix[4] < sz[4]; ix[4]++)
    for (ix[3] = 0; ix[3] < sz[3]; ix[3]++)
      for (ix[2] = 0; ix[2] < sz[2]; ix[2]++)
        for (ix[1] = 0; ix[1] < sz[1]; ix[1]++)
          for (ix[0] = 0; ix[0] < sz[0]; ix[0]++)
            { posA[kpos] = ix_position(A->d, ix, A->b, A->s);
              posB[kpos] = ix_position(B->d, ix, B->b, B->s);
              posC[kpos] = ix_position(C->d, ix, C->b, C->s);
              kpos++;
            }
  assert(kpos == npos);
  if (verbose) { fprintf(stderr, "  counted %llu tuples\n", npos); }
  
  /* Check forward enumeration, Fortran order: */
  kpos = 0;
  ix_enum(check_fw, d, sz, ix_order_F, FALSE, A->b, A->s, B->b, B->s, C->b, C->s);
  assert(kpos == npos);
  
  /* Check backward enumeration, Fortran order: */
  kpos = npos;
  ix_enum(check_bw, d, sz, ix_order_F, TRUE,  A->b, A->s, B->b, B->s, C->b, C->s);
  assert(kpos == 0);
  
  /* Enumerate all positions by hand in "L" order: */
  kpos = 0;
  for (ix[0] = 0; ix[0] < sz[0]; ix[0]++)
    for (ix[1] = 0; ix[1] < sz[1]; ix[1]++)
      for (ix[2] = 0; ix[2] < sz[2]; ix[2]++)
        for (ix[3] = 0; ix[3] < sz[3]; ix[3]++)
          for (ix[4] = 0; ix[4] < sz[4]; ix[4]++)
            { posA[kpos] = ix_position(A->d, ix, A->b, A->s);
              posB[kpos] = ix_position(B->d, ix, B->b, B->s);
              posC[kpos] = ix_position(C->d, ix, C->b, C->s);
              kpos++;
            }
  assert(kpos == npos);
  
  /* Check forward enumeration, C/Pascal order: */
  kpos = 0;
  ix_enum(check_fw, d, sz, ix_order_L, FALSE, A->b, A->s, B->b, B->s, C->b, C->s);
  assert(kpos == npos);
  
  /* Check backward enumeration, C/Pascal order: */
  kpos = npos;
  ix_enum(check_bw, d, sz, ix_order_L, TRUE,  A->b, A->s, B->b, B->s, C->b, C->s);
  assert(kpos == 0);

  /* IMPLEMENTATION OF LOCAL PROCS */

  void check_fw( ix_index_t ix[], ix_pos_t pA, ix_pos_t pB, ix_pos_t pC )
    {
      assert(kpos >= 0);
      assert(kpos < npos);
      assert(pA == posA[kpos]);
      assert(pB == posB[kpos]);
      assert(pC == posC[kpos]);
      kpos++;
    }
  
  void check_bw( ix_index_t ix[], ix_pos_t pA, ix_pos_t pB, ix_pos_t pC )
    {
      kpos--;
      assert(kpos >= 0);
      assert(kpos < npos);
      assert(pA == posA[kpos]);
      assert(pB == posB[kpos]);
      assert(pC == posC[kpos]);
    }
}

void test_packed(desc_t *A)
{
  fprintf(stderr, "Checking ix_packed_XXX ops...\n");
  
  /* Fabricate a *packed* array (can't use {A}, but use its size): */
  ix_dim_t d = A->d;
  ix_size_t *sz = A->z;
  desc_t *F = make_packed_desc(d, sz, ix_order_F);
  desc_t *L = make_packed_desc(d, sz, ix_order_L);
  bool_t verbose = FALSE;
  if (verbose) 
    { print_desc(stderr, "A = { ", A, " }\n");
      print_desc(stderr, "F = { ", F, " }\n");
      print_desc(stderr, "L = { ", L, " }\n");
    }
    
  ix_index_t ix[MAXDIM], jx[MAXDIM];
  int i;

  /* Allocate a vector to store all positions of {B}: */
  ix_pos_count_t npos = ix_num_tuples(d, sz);
  int64_t kpos;     /* Current index into {posA}. */

  /* Enumerate all tuples by hand in "F" order, check {F} positions: */
  kpos = 0;
  for (ix[4] = 0; ix[4] < sz[4]; ix[4]++)
    for (ix[3] = 0; ix[3] < sz[3]; ix[3]++)
      for (ix[2] = 0; ix[2] < sz[2]; ix[2]++)
        for (ix[1] = 0; ix[1] < sz[1]; ix[1]++)
          for (ix[0] = 0; ix[0] < sz[0]; ix[0]++)
            { ix_pos_t posF1 = ix_packed_position(d, ix, sz, ix_order_F);
              ix_pos_t posF2 = ix_position(d, ix, F->b, F->s);
              if (verbose)
                { fprintf(stderr, "  kpos = %llu\n", kpos);  
                  fprintf(stderr, "  posF1 = %llu\n", posF1); 
                  fprintf(stderr, "  posF2 = %llu\n", posF2);
                }
              assert(posF1 == kpos);
              assert(posF2 == posF1);
              ix_packed_indices(d, kpos, sz, ix_order_F, jx);
              for (i = 0; i < d; i++) { assert(ix[i] == jx[i]); }
              kpos++;
            }
  assert(kpos == npos);

  /* Enumerate all tuples by hand in "L" order, check {L} positions: */
  kpos = 0;
  for (ix[0] = 0; ix[0] < sz[0]; ix[0]++)
    for (ix[1] = 0; ix[1] < sz[1]; ix[1]++)
      for (ix[2] = 0; ix[2] < sz[2]; ix[2]++)
        for (ix[3] = 0; ix[3] < sz[3]; ix[3]++)
          for (ix[4] = 0; ix[4] < sz[4]; ix[4]++)
            { ix_pos_t posL1 = ix_packed_position(d, ix, sz, ix_order_L);
              ix_pos_t posL2 = ix_position(d, ix, L->b, L->s);
              assert(posL1 == kpos);
              assert(posL2 == posL1);
              ix_packed_indices(d, kpos, sz, ix_order_L, jx);
              for (i = 0; i < d; i++) { assert(ix[i] == jx[i]); }
              kpos++;
            }
  assert(kpos == npos);

}

void test_crop(desc_t *A)
{
  fprintf(stderr, "Checking ix_crop...\n");
  bool_t verbose = FALSE;
  if (verbose) { print_desc(stderr, "A = { ", A, " }\n"); } 
  int trial, ntrials = 40;
  ix_index_t sh[MAXDIM]; /* Shift of cropped array along each axis. */
  ix_size_t sz[MAXDIM]; /* Size of cropped array along each axis. */
  for (trial = 0; trial < ntrials; trial++)
    { /* Initialze {sh} and {sz} for full array: */
      sh[0] = sh[1] = sh[2] = sh[3] = sh[4] = 0;
      { int i; for (i = 0; i < A->d; i++) { sz[i] = A->z[i]; } }
      /* Start with the standard array: */
      desc_t B = (*A);
      if (verbose) { print_desc(stderr, "B = { ", &B, " }\n"); } 

      /* Perform first cropping: */
      ix_axis_t ax1 = (trial % A->d);
      ix_size_t skip1 = (trial & 3)*sz[ax1]/8;
      ix_size_t keep1 = (trial/4 & 3)*(sz[ax1] - skip1)/8;
      if (keep1 < 0) { keep1 = 0; }
      sh[ax1] += skip1; sz[ax1] = keep1;
      ix_crop(B.d, B.z, &B.b, B.s, ax1, skip1, keep1);
      if (verbose) 
        { fprintf(stderr, "crop(%d, %llu, %llu)", ax1, skip1, keep1);
          print_desc(stderr, " = { ", &B, " }\n");
        } 
      /* Check validity of descriptor: */
      if (! ix_parms_are_valid(B.d, B.z, B.b, B.s, FALSE))
        { bug("ix_crop returns an invalid desc"); }
      /* Check whether the size of {B} is correct: */
      check_sizes(B.d, B.z, sz);
      
      /* Perform second cropping (possibly on same axis): */ 
      ix_axis_t ax2 = ((trial / A->d) % A->d);
      ix_size_t skip2 = (trial/2 & 3)*sz[ax2]/8;
      ix_size_t keep2 = (trial/5 & 3)*(sz[ax2] - skip2)/8;
      if (keep2 < 0) { keep2 = 0; }
      sh[ax2] += skip2; sz[ax2] = keep2;
      ix_crop(B.d, B.z, &B.b, B.s, ax2, skip2, keep2);
      if (verbose) 
        { fprintf(stderr, "crop(%d, %llu, %llu)", ax2, skip2, keep2);
          print_desc(stderr, " = { ", &B, " }\n");
        } 
      /* Check validity of descriptor: */
      if (! ix_parms_are_valid(B.d, B.z, B.b, B.s, FALSE))
        { bug("ix_crop returns an invalid desc"); }
      
      /* Check whether the size of {B} is correct: */
      check_sizes(B.d, B.z, sz);
      /* Now check coincidence of the two arrays: */
      ix_index_t ixA[MAXDIM], ixB[MAXDIM];
      for (ixB[4] = 0; ixB[4] < sz[4]; ixB[4]++)
        for (ixB[3] = 0; ixB[3] < sz[3]; ixB[3]++)
          for (ixB[2] = 0; ixB[2] < sz[2]; ixB[2]++)
            for (ixB[1] = 0; ixB[1] < sz[1]; ixB[1]++)
              for (ixB[0] = 0; ixB[0] < sz[0]; ixB[0]++)
                { 
                  ix_assign(A->d, ixA, ixB);
                  ix_shift(A->d, ixA, sh);
                  ix_pos_t addrB = ix_position(B.d, ixB, B.b, B.s);
                  ix_pos_t addrA = ix_position(A->d, ixA, A->b, A->s);
                  if (addrB != addrA)
                    { bug("crop address error addrB = %llu addrA = %llu", addrB, addrA); } 
                }
    }
}

void test_subsample(desc_t *A) 
{
  fprintf(stderr, "Checking ix_subsample...\n");
  bool_t verbose = FALSE;
  if (verbose) { print_desc(stderr, "A = { ", A, " }\n"); } 
  int trial, ntrials = 40;
  ix_size_t st[MAXDIM]; /* Subsampling step along each axis. */
  ix_size_t sz[MAXDIM]; /* Size of subsampled array along each axis. */
  for (trial = 0; trial < ntrials; trial++)
    { /* Initialze {st} and {sz} for full array: */
      st[0] = st[1] = st[2] = st[3] = st[4] = 1;
      { int i; for (i = 0; i < A->d; i++) { sz[i] = A->z[i]; } }
      /* Start with the standard array: */
      desc_t B = (*A);
      
      /* Perform first subsampling: */
      ix_axis_t ax1 = (trial % A->d);
      ix_size_t step1 = 1 + (trial & 3)*(sz[ax1] - 1)/8;
      ix_size_t keep1 = (sz[ax1] + step1 - 1)/step1;
      st[ax1] *= step1; sz[ax1] = keep1;
      ix_subsample(B.d, B.z, &(B.b), B.s, ax1, step1);
      if (verbose) 
        { fprintf(stderr, "subsample(%d, %llu)", ax1, step1);
          print_desc(stderr, " = { ", &B, " }\n");
        } 
      /* Check validity of descriptor: */
      if (! ix_parms_are_valid(B.d, B.z, B.b, B.s, FALSE)) 
        { bug("ix_flip returns an invalid desc"); }
      
      /* Perform second subsampling (addrsibly on same axis): */ 
      ix_axis_t ax2 = ((trial / A->d) % A->d);
      ix_size_t step2 = 1 + (trial/2 & 3)*(sz[ax2] - 1)/8;
      ix_size_t keep2 = (sz[ax2] + step2 - 1)/step2;
      st[ax2] *= step2; sz[ax2] = keep2;
      ix_subsample(B.d, B.z, &(B.b), B.s, ax2, step2);
      if (verbose) 
        { fprintf(stderr, "subsample(%d, %llu)", ax2, step2);
          print_desc(stderr, " = { ", &B, " }\n");
        } 
      /* Check validity of descriptor: */
      if (! ix_parms_are_valid(B.d, B.z, B.b, B.s, FALSE)) 
        { bug("ix_subsample returns an invalid desc"); }
      
      check_sizes(B.d, B.z, sz); 
      /* Now check coincidence of the two arrays: */
      ix_index_t ixA[MAXDIM], ixB[MAXDIM];
      for (ixB[4] = 0; ixB[4] < sz[4]; ixB[4]++)
        for (ixB[3] = 0; ixB[3] < sz[3]; ixB[3]++)
          for (ixB[2] = 0; ixB[2] < sz[2]; ixB[2]++)
            for (ixB[1] = 0; ixB[1] < sz[1]; ixB[1]++)
              for (ixB[0] = 0; ixB[0] < sz[0]; ixB[0]++)
                { 
                  ix_axis_t ax;
                  for (ax = 0; ax < A->d; ax++) { ixA[ax] = ixB[ax]*st[ax]; }
                  ix_pos_t addrB = ix_position(B.d, ixB, B.b, B.s);
                  ix_pos_t addrA = ix_position(A->d, ixA, A->b, A->s);
                  if (addrB != addrA)
                    { bug("subsample address error addrB = %llu addrA = %llu", addrB, addrA); } 
                }
    }
}

void test_swap_indices(desc_t *A)
{
  fprintf(stderr, "Checking ix_swap_indices...\n");
  bool_t verbose = FALSE;
  if (verbose) { print_desc(stderr, "A = { ", A, " }\n"); } 
  int trial, ntrials = 50;
  ix_axis_t tr[MAXDIM]; /* Axis {i} of {B} is axis {tr[i]} of {A}. */
  for (trial = 0; trial < ntrials; trial++)
    { /* Initialze {st} and {sz} for full array: */
      tr[0] = 0; tr[1] = 1; tr[2] = 2; tr[3] = 3; tr[4] = 4; 
      /* Start with the standard array: */
      desc_t B = (*A);

      /* Perform first transaddress: */
      ix_axis_t ax1 = (trial % A->d), bx1 = (trial/2 % A->d);
      { ix_axis_t tmp = tr[ax1]; tr[ax1] = tr[bx1]; tr[bx1] = tmp; }
      ix_swap_indices(B.d, B.z, &(B.b), B.s, ax1, bx1);
      if (verbose) 
        { fprintf(stderr, "swap_indices(%d, %d)", ax1, bx1);
          print_desc(stderr, " = { ", &B, " }\n");
        } 
      /* Check validity of descriptor: */
      if (! ix_parms_are_valid(B.d, B.z, B.b, B.s, FALSE))
        { bug("ix_flip returns an invalid desc"); }

      /* Perform second transaddress (possibly on same axis): */ 
      ix_axis_t ax2 = ((trial/A->d) % A->d), bx2 = ((trial/10) % A->d);
      { ix_axis_t tmp = tr[ax2]; tr[ax2] = tr[bx2]; tr[bx2] = tmp; }
      ix_swap_indices(B.d, B.z, &(B.b), B.s, ax2, bx2);
      if (verbose) 
        { fprintf(stderr, "swap_indices(%d, %d)", ax2, bx2);
          print_desc(stderr, " = { ", &B, " }\n");
        } 
      /* Check validity of descriptor: */
      if (! ix_parms_are_valid(B.d, B.z, B.b, B.s, FALSE))
        { bug("ix_swap_indices returns an invalid desc"); }
      
      /* Check whether the size and increments of {B} is correct: */
      { int i;
        for (i = 0; i < A->d; i++)
          { if (B.z[i] != A->z[tr[i]])
              { bug("B.z[%d] = %llu  A.z[%d] = %llu", i, B.z[i], tr[i], A->z[tr[i]]); }
            if (B.s[0] != A->s[tr[0]])
              { bug("B.s[%d] = %llu  A.s[%d] = %llu", i, B.s[i], tr[i], A->s[tr[i]]); }
          }
      }
      /* Now check coincidence of the two arrays: */
      ix_index_t ixA[MAXDIM], ixB[MAXDIM];
      for (ixA[4] = 0; ixA[4] < A->z[4]; ixA[4]++)
        for (ixA[3] = 0; ixA[3] < A->z[3]; ixA[3]++)
          for (ixA[2] = 0; ixA[2] < A->z[2]; ixA[2]++)
            for (ixA[1] = 0; ixA[1] < A->z[1]; ixA[1]++)
              for (ixA[0] = 0; ixA[0] < A->z[0]; ixA[0]++)
                { 
                  ix_axis_t ax;
                  for (ax = 0; ax < A->d; ax++) { ixB[ax] = ixA[tr[ax]]; }
                  ix_pos_t addrB = ix_position(B.d, ixB, B.b, B.s);
                  ix_pos_t addrA = ix_position(A->d, ixA, A->b, A->s);
                  if (addrB != addrA)
                    { bug("swap_indices address error addrB = %llu addrA = %llu", addrB, addrA); } 
                }
    }
}

void test_flip_indices(desc_t *A)
{
  fprintf(stderr, "NOT Checking ix_flip_indices...\n");
}

void test_flip(desc_t *A)
{
  fprintf(stderr, "Checking ix_flip...\n");
  bool_t verbose = FALSE;
  if (verbose) { print_desc(stderr, "A = { ", A, " }\n"); } 
  int trial, ntrials = 40;
  bool_t fp[MAXDIM]; /* Tells whether each axis was flipped or not. */
  for (trial = 0; trial < ntrials; trial++)
    { /* Initialze {fp} for unflipped array: */
      fp[0] = fp[1] = fp[2] = fp[3] = fp[4] = FALSE;
      /* Start with the standard array: */
      desc_t B = (*A);

      /* Perform first flip: */
      ix_axis_t ax1 = (trial % A->d);
      fp[ax1] = ! fp[ax1];
      ix_flip(B.d, B.z, &(B.b), B.s, ax1);
      if (verbose) 
        { fprintf(stderr, "flip(%d)", ax1);
          print_desc(stderr, " = { ", &B, " }\n");
        } 
      /* Check validity of descriptor: */
      if (! ix_parms_are_valid(B.d, B.z, B.b, B.s, FALSE))
        { bug("ix_flip returns an invalid desc"); }

      /* Perform second flip (possibly on same axis): */ 
      ix_axis_t ax2 = ((trial / A->d) % A->d);
      fp[ax2] = ! fp[ax2];
      ix_flip(B.d, B.z, &(B.b), B.s, ax2);
      if (verbose) 
        { fprintf(stderr, "flip(%d)", ax2);
          print_desc(stderr, " = { ", &B, " }\n");
        } 
      /* Check validity of descriptor: */
      if (! ix_parms_are_valid(B.d, B.z, B.b, B.s, FALSE))
        { bug("ix_flip returns an invalid desc"); }

      /* Check whether the size of {B} is correct: */
      check_sizes(B.d, B.z, A->z);
      /* Check whether the increments of {B} are correct: */
      if (B.s[0] != A->s[0]*(fp[0]?-1:+1)) { bug("step[0] = %llu", B.s[0]); }
      if (B.s[1] != A->s[1]*(fp[1]?-1:+1)) { bug("step[1] = %llu", B.s[1]); }
      if (B.s[2] != A->s[2]*(fp[2]?-1:+1)) { bug("step[2] = %llu", B.s[2]); }
      if (B.s[3] != A->s[3]*(fp[3]?-1:+1)) { bug("step[3] = %llu", B.s[3]); }
      if (B.s[4] != A->s[4]*(fp[4]?-1:+1)) { bug("step[4] = %llu", B.s[4]); }
      /* Now check coincidence of the two arrays: */
      ix_index_t ixA[MAXDIM], ixB[MAXDIM];
      for (ixB[4] = 0; ixB[4] < B.z[4]; ixB[4]++)
        for (ixB[3] = 0; ixB[3] < B.z[3]; ixB[3]++)
          for (ixB[2] = 0; ixB[2] < B.z[2]; ixB[2]++)
            for (ixB[1] = 0; ixB[1] < B.z[1]; ixB[1]++)
              for (ixB[0] = 0; ixB[0] < B.z[0]; ixB[0]++)
                { 
                  ix_axis_t ax;
                  for (ax = 0; ax < A->d; ax++) 
                    { ixA[ax] = (fp[ax] ? A->z[ax] - 1 - ixB[ax] : ixB[ax]); }
                  ix_pos_t addrB = ix_position(B.d, ixB, B.b, B.s);
                  ix_pos_t addrA = ix_position(A->d, ixA, A->b, A->s);
                  if (addrB != addrA)
                    { bug("flip address error addrB = %llu addrA = %llu", addrB, addrA); } 
                }
    }
}

void test_diagonal(desc_t *A)
{
  fprintf(stderr, "Checking ix_diagonal...\n");
  bool_t verbose = FALSE;
  if (verbose) { print_desc(stderr, "A = { ", A, " }\n"); } 
  int trial, ntrials = 50;
  ix_size_t sz[MAXDIM]; /* Size of diagonalized array along each axis. */
  for (trial = 0; trial < ntrials; trial++)
    { /* Initialze {sz} for full array: */
      { int i; for (i = 0; i < A->d; i++) { sz[i] = A->z[i]; } }
      /* Start with the standard array: */
      desc_t B = (*A);

      /* Choose any two axes: */
      ix_axis_t ax1 = (trial % A->d), bx1 = (trial/A->d % A->d);
      if (bx1 == ax1) { bx1 = (ax1 + 1) % A->d; }
      /* Make it square: */
      ix_size_t szmin = (sz[ax1] < sz[bx1] ? sz[ax1] : sz[bx1]);
      ix_crop(B.d, B.z, &(B.b), B.s, ax1, 0, szmin); sz[ax1] = szmin;
      ix_crop(B.d, B.z, &(B.b), B.s, bx1, 0, szmin); sz[bx1] = szmin;
      /* Diagonalize array: */
      ix_diagonal(B.d, B.z, &(B.b), B.s, ax1, bx1);
      /* Compute expected sizes of {1}: */
      if (sz[bx1] < sz[ax1]) { sz[ax1] = sz[bx1]; }
      if (bx1 != ax1) { sz[bx1] = 1; }
      if (verbose) 
        { fprintf(stderr, "diagonal(%d, %d)", ax1, bx1);
          print_desc(stderr, " = { ", &B, " }\n");
        } 
      /* Check validity of descriptor: */
      if (! ix_parms_are_valid(B.d, B.z, B.b, B.s, FALSE))
        { bug("ix_diagonal returns an invalid desc"); }

      /* Check whether the size of {1} is correct: */
      check_sizes(B.d, B.z, sz);
      /* Now check coincidence of the two arrays: */
      ix_index_t ixA[MAXDIM], ixB[MAXDIM];
      for (ixB[4] = 0; ixB[4] < B.z[4]; ixB[4]++)
        for (ixB[3] = 0; ixB[3] < B.z[3]; ixB[3]++)
          for (ixB[2] = 0; ixB[2] < B.z[2]; ixB[2]++)
            for (ixB[1] = 0; ixB[1] < B.z[1]; ixB[1]++)
              for (ixB[0] = 0; ixB[0] < B.z[0]; ixB[0]++)
                { 
                  ix_assign(A->d, ixA, ixB);
                  ixA[bx1] = ixA[ax1];
                  ix_pos_t addrB = ix_position(B.d, ixB, B.b, B.s);
                  ix_pos_t addrA = ix_position(A->d, ixA, A->b, A->s);
                  if (addrB != addrA)
                    { bug("diagonal address error addrB = %llu addrA = %llu", addrB, addrA); } 
                }
    }
}

void test_chop(desc_t *A)
{
  fprintf(stderr, "Checking ix_chop...\n");
  bool_t verbose = FALSE;
  if (verbose) { print_desc(stderr, "A = { ", A, " }\n"); } 
  int trial, ntrials = 50;
  ix_size_t sz[MAXDIM]; /* Size of chopped array along each axis. */
  for (trial = 0; trial < ntrials; trial++)
    { /* Initialze {sz} for full array: */
      { int i; for (i = 0; i < A->d; i++) { sz[i] = A->z[i]; } }
      /* Start with the standard array: */
      desc_t B = (*A);

      /* Choose two DISTINCT axes: */
      ix_axis_t ax1 = (trial % A->d);
      ix_axis_t bx1 = (trial/A->d % (A->d - 1));
      if (bx1 >= ax1) { bx1++; }
      /* Choose a nonzero chunk size: */
      ix_size_t chunksz1 = 1 + (trial/3 & 3)*(A->z[ax1]-1)/8;
      /* Ensure integral number of chunks: */
      ix_size_t szround = (sz[ax1]/chunksz1)*chunksz1;
      ix_crop(B.d, B.z, &(B.b), B.s, ax1, 0, szround); sz[ax1] = szround;
      /* Ensure that the argument of {chop} is trivial along axis {bx1}: */
      ix_crop(B.d, B.z, &(B.b), B.s, bx1, 0, 1); sz[bx1] = 1;
      /* Chop array: */
      ix_chop(B.d, B.z, &(B.b), B.s, ax1, chunksz1, bx1);
      /* Compute expected sizes of {1}: */
      sz[bx1] = sz[ax1]/chunksz1; sz[ax1] = chunksz1;
      if (verbose) 
        { fprintf(stderr, "chop(%d, %lld, %d)", ax1, chunksz1, bx1);
          print_desc(stderr, " = { ", &B, " }\n");
        } 
      /* Check validity of descriptor: */
      if (! ix_parms_are_valid(B.d, B.z, B.b, B.s, FALSE))
        { bug("ix_chop returns an invalid desc"); }

      /* Check whether the size of {1} is correct: */
      check_sizes(B.d, B.z, sz);
      /* Now check coincidence of the two arrays: */
      ix_index_t ixA[MAXDIM], ixB[MAXDIM];
      for (ixB[4] = 0; ixB[4] < B.z[4]; ixB[4]++)
        for (ixB[3] = 0; ixB[3] < B.z[3]; ixB[3]++)
          for (ixB[2] = 0; ixB[2] < B.z[2]; ixB[2]++)
            for (ixB[1] = 0; ixB[1] < B.z[1]; ixB[1]++)
              for (ixB[0] = 0; ixB[0] < B.z[0]; ixB[0]++)
                { 
                  ix_assign(A->d, ixA, ixB);
                  ixA[ax1] = ixA[ax1] + chunksz1 * ixB[bx1];
                  ixA[bx1] = 0;
                  ix_pos_t addrB = ix_position(B.d, ixB, B.b, B.s);
                  ix_pos_t addrA = ix_position(A->d, ixA, A->b, A->s);
                  if (addrB != addrA)
                    { bug("chop address error addrB = %llu addrA = %llu", addrB, addrA); } 
                }
    }
}
  
void test_replicate(desc_t *A)
{
  fprintf(stderr, "Checking ix_replicate...\n");
  bool_t verbose = FALSE;
  if (verbose) { print_desc(stderr, "A = { ", A, " }\n"); } 
  int trial, ntrials = 50;
  ix_size_t sz[MAXDIM]; /* Size of replicated array along each axis. */
  for (trial = 0; trial < ntrials; trial++)
    { /* Initialze {sz} for full array: */
      { int i; for (i = 0; i < A->d; i++) { sz[i] = A->z[i]; } }
      /* Start with the standard array: */
      desc_t B = (*A);
      
      /* Choose axis for first replication: */
      ix_axis_t ax1 = (trial % A->d);
      /* Choose a positive replication factor: */
      ix_size_t rep1 = 1 + (trial/3 & 3)*A->d;
      /* Make trivial along axis {ax1}: */
      ix_size_t skip1 = (B.z[ax1]-1)/2;
      ix_crop(B.d, B.z, &(B.b), B.s, ax1, skip1, 1); sz[ax1] = 1;
      /* Replicate along axis {ax1}: */
      ix_replicate(B.d, B.z, &(B.b), B.s, ax1, rep1);
      sz[ax1] = rep1;
      if (verbose) 
        { fprintf(stderr, "chop(%d,%llu,1)+replicate(%d,%lld)", ax1, skip1, ax1, rep1);
          print_desc(stderr, " = { ", &B, " }\n");
        } 
      /* Check validity of descriptor: */
      if (! ix_parms_are_valid(B.d, B.z, B.b, B.s, FALSE))
        { bug("ix_replicate returns an invalid desc"); }
      
      /* Choose axis for second replication (may be {ax1}): */
      ix_axis_t ax2 = (trial/A->d % A->d);
      /* Choose a replication factor: */
      ix_size_t rep2 = 1 + (trial/8 & 3)*A->d;
      /* Make trivial along axis {ax2}: */
      ix_size_t skip2 = (B.z[ax2]-1)/2;
      ix_crop(B.d, B.z, &(B.b), B.s, ax2, skip2, 1); sz[ax2] = 1;
      /* Replicate along axis {ax2}: */
      ix_replicate(B.d, B.z, &(B.b), B.s, ax2, rep2);
      sz[ax2] = rep2;
      if (verbose) 
        { fprintf(stderr, "chop(%d,%llu,1)+replicate(%d,%lld)", ax2, skip2, ax2, rep2);
          print_desc(stderr, " = { ", &B, " }\n");
        } 
      /* Check validity of descriptor: */
      if (! ix_parms_are_valid(B.d, B.z, B.b, B.s, FALSE))
        { bug("ix_replicate returns an invalid desc"); }

      /* Check whether the size of {1} is correct: */
      check_sizes(B.d, B.z, sz);
      /* Now check coincidence of the two arrays: */
      ix_index_t ixA[MAXDIM], ixB[MAXDIM];
      for (ixB[4] = 0; ixB[4] < B.z[4]; ixB[4]++)
        for (ixB[3] = 0; ixB[3] < B.z[3]; ixB[3]++)
          for (ixB[2] = 0; ixB[2] < B.z[2]; ixB[2]++)
            for (ixB[1] = 0; ixB[1] < B.z[1]; ixB[1]++)
              for (ixB[0] = 0; ixB[0] < B.z[0]; ixB[0]++)
                { 
                  ix_assign(A->d, ixA, ixB);
                  ixA[ax1] = skip1;
                  /* The second crop changes {b} only if {ax1 != ax2}: */
                  if (ax2 != ax1) { ixA[ax2] = skip2; }
                  ix_pos_t addrB = ix_position(B.d, ixB, B.b, B.s);
                  ix_pos_t addrA = ix_position(A->d, ixA, A->b, A->s);
                  if (addrB != addrA)
                    { bug("replicate address error addrB = %llu addrA = %llu", addrB, addrA); } 
                }
    }
}
  
void check_max_indices(ix_dim_t d, ix_index_t *ma, ix_index_t *mb)
{
  ix_axis_t i;
  for (i = 0; i < d; i++)
    { if (ma[i] != mb[i]) 
        { bug("(ma[%d] = %lld mb[%d] = %lld", i, ma[i], i, mb[i]); }
    }
}
  
void check_max_indices_sz(ix_dim_t d, ix_index_t *ma, ix_size_t *szb)
{
  ix_axis_t i;
  for (i = 0; i < d; i++)
    { ix_index_t mbi = szb[i]-1;
      if (ma[i] != mbi) 
        { bug("(ma[%d]-1) = %lld mb[%d] = %lld", i, ma[i], i, mbi); }
    }
}
  
void check_sizes(ix_dim_t d, ix_size_t *sza, ix_size_t *szb)
{
  ix_axis_t i;
  for (i = 0; i < d; i++)
    { if (sza[i] != szb[i]) 
        { bug("sza[%d] = %llu szb[%d] = %llu", i, sza[i], i, szb[i]); }
    }
}

void print_desc(FILE *wr, char *pf, desc_t *A, char *sf)
  { fprintf(stderr, "%s", pf); 
    int i;
    for (i = 0; i < A->d; i++)
      { fprintf(stderr, " %llu(%lld)", A->z[i], A->s[i]); }
    fprintf(stderr, 
      " b = %llu tuples = %llu addresses = %llu", 
      A->b, 
      ix_num_tuples(A->d,A->z), 
      ix_num_positions(A->d,A->z,A->s)
    );
    fprintf(stderr, "%s", sf);
  }
