/* See dm_filter.h */
/* Last edited on 2008-04-15 10:33:12 by stolfi */

#define dm_filter_C_COPYRIGHT \
  "Copyright  2005  by the State University of Campinas (UNICAMP)" \
  " and the Federal Fluminense University (UFF)"

#include <dm_filter.h>
#include <dm_nucleic.h>
#include <dm_basic.h>
#include <dm_sample.h>
#include <dm_datum.h>
#include <dm_seq.h>

#include <msm_image.h>
#include <msm_rung.h>
#include <msm_pairing.h>
#include <msm_cand.h>

#include <vec.h>
#include <nget.h>
#include <fget.h>
#include <filefmt.h>
#include <jsrandom.h>
#include <pswr.h>

#include <fftw3.h>

#include <math.h>
#include <stdlib.h>
#include <stdio.h>
#include <ctype.h>
#include <assert.h>

/* IMPLEMENTATIONS */
 
int dm_filter_output_num_sample_datums(int nold, bool_t circ, int nwtb)
  { return (nold - (circ ? 0 : nwtb - 1))/2; }

dm_seq_t dm_filter_seq(dm_seq_t *oldp, double_vec_t *wtbp, char *wcmt, int nsub)
  { int nold = oldp->dv.ne;
    int nwtb = wtbp->ne;
    int nnew = dm_filter_output_num_sample_datums(nold, oldp->sd.circ, nwtb);
    if (nnew < 0) { nnew = 0; }
    /* Allocate new datum vector: */
    dm_datum_vec_t new_dv = dm_datum_vec_new(nnew);
    /* Scale factors for new samples: */
    dm_datum_scale_t new_sfac;
    int k;
    if (nnew == 0)
      { /* Sequence is too short to be filtered. */
        /* Set {new_sfac} to something sensible, just in case: */
        for (k = 0; k < dm_CHANNELS; k++) { new_sfac.f[k] = oldp->sfac.f[k]/M_SQRT2; }
        /* Vector {new_dv} is empty. */
      }
    else
      { 
        if (! oldp->sd.circ) { assert(2*(nnew-1) + (nwtb-1) < nold); }

        /* Process datum vector, one channel at a time; */
        double_vec_t old_v = double_vec_new(nold); /* Decoded samples of old sequence. */
        double_vec_t new_v = double_vec_new(nnew); /* Decoded samples of new sequence. */
        for (k = 0; k < dm_CHANNELS; k++)
          { /* Extract and decode the old samples {old_v}: */
            double old_fac = oldp->sfac.f[k];
            int i;
            for (i = 0; i < nold; i++)
              { dm_datum_t *dp = dm_seq_get_datum_address(oldp, i);
                dm_sample_t s = dp->c[k];
                old_v.e[i] = dm_sample_decode(s)*old_fac; 
              }

            /* Filter the old samples to produce the new samples {new_v}: */
            dm_double_vec_filter(&old_v, wtbp, &new_v);

            /* Compute the new scale factor {new_fac}: */
            double sum_sq = 0;
            for (i = 0; i < nnew; i++) { double v = new_v.e[i]; sum_sq += v*v; }
            new_sfac.f[k] = sqrt(sum_sq/nnew);
            /* fprintf(stderr, "  channel %d sum_sq = %14.8f  sfac = %12.8f\n", k, sum_sq, new_sfac.f[k]); */

            /* Encode the new samples and store them into the new datum vector: */
            for (i = 0; i < nnew; i++)
              { dm_sample_t s = dm_sample_encode(new_v.e[i]/new_sfac.f[k]);
                dm_datum_t *dp = &(new_dv.e[i]);
                dp->c[k] = s; 
              }
          }

      }
      
    /* Append the weight table comment to the sequence's comment string: */
    char *new_cmt = NULL;
    if (wcmt == NULL) { wcmt = "unnamed table"; }
    char *old_cmt = (oldp->cmt != NULL ? oldp->cmt : "");
    asprintf(&(new_cmt), "%sfiltered with %s (window width %d)\n", old_cmt, wcmt, nwtb);

    /* Pack the filtered datum vector as a {dm_seq_t}: */
    dm_seq_t new = dm_seq_from_datum_vec
      ( oldp->sd.id, oldp->sd.name, oldp->sd.level+1, oldp->sd.circ, 
        new_cmt, &new_sfac, new_dv, nsub
      );
    return new;
  }

void dm_double_vec_filter(double_vec_t *oldp, double_vec_t *wtbp, double_vec_t *newp)
  { int nold = oldp->ne;
    int nwtb = wtbp->ne;
    int nnew = newp->ne;
    int i, j;
    for (i = 0; i < nnew; i++)
      { double sum_w = 0, sum_ws = 0;
        for(j = 0; j < nwtb; j++)
          { double w = wtbp->e[j]; 
            double s = oldp->e[(2*i+j) % nold];
            sum_w += w; 
            sum_ws += w*s; 
          }
        newp->e[i] = sum_ws/sum_w;
      }
  }

void dm_filter_seq_multi
  ( dm_seq_t *s, 
    int maxLevel, 
    double_vec_t *wtb0,
    char *wname0,
    double_vec_t *wtb1,
    char *wname1,
    dm_seq_t sr[],
    int nsub
  )
  { int level;
    for (level = 0; level <= maxLevel; level++)
      { if (level == 0)
          { sr[level] = dm_seq_copy_datums(s); }
        else
          { double_vec_t *wtb = (level == 1 ? wtb0 : wtb1);
            char *wname = (level == 1 ? wname0 : wname1);
            sr[level] = dm_filter_seq(&(sr[level-1]), wtb, wname, nsub);
          }
      }
  }

