#ifndef dm_score_H
#define dm_score_H

/* Scoring functions for rungs and steps of pairings. */
/* Last edited on 2008-01-29 18:59:47 by hcgl */

#define dm_score_H_COPYRIGHT \
  "Copyright  2006  by the State University of Campinas (UNICAMP)" \
  " and the Federal Fluminense University (UFF)"

#include <dm_basic.h>
#include <dm_datum.h>
#include <dm_seq.h>

#include <msm_rung.h>

#include <argparser.h>

typedef struct dm_score_args_t
  { /* Data-related parameters: */
    double eql;      /* Score increment (us. positive) for a rung between equal raw datums. */
    double dif;      /* Score increment (us. negative) for a rung between distinct raw datums. */
    /* Move-related parameters: */
    double brk;      /* Score increment (us. negative) for imperfect steps. */
    double skp;      /* Score increment (us. negative) for skipping a datum pair. */
    double unp;      /* Score increment (us. negative) for each unpaired datum. */
  } dm_score_args_t; 
  /* A scoring criterion for rungs and steps between sequences.
  
    A {dm_score_args_t} record defines a numerical score for any rung
    {g} and any step {g0->g1} between two datum sequences {x} and {y}.
    By extension, it also defines a score for any pairing {pr} between
    the two sequences, which is simply the sum of the scores of all
    its fundamental steps.
    
    The scoring parameters are typically chosen so that the score is
    positive when the paired datums are very similar, and negative
    when they are very dissimilar. */

/* RUNG AND STEP SCORES */

double dm_score_rung
  ( dm_score_args_t *sc, 
    msm_rung_t g,
    dm_seq_t *xp,
    dm_seq_t *yp
  );
  /* Computes the score of a rung {g} between sequences {*xp} and
    {*yp}, according to the scoring criterion {sc}.
    
    The result depends on the weights {sc.eql} and {sc.dif} and the
    datums {dx} of {xp} and {dy} of {yp} that are associated by {g}. */

double dm_score_step
  ( dm_score_args_t *sc, 
    msm_rung_t g0, 
    msm_rung_t g1, 
    dm_seq_t *xp, 
    dm_seq_t *yp
  );
  /* Computes the score of a step {g0-->g1} between sequences {*xp} and
    {*yp}, according to the scoring criterion {sc}.  The step must
    be 1-increasing.
    
    The result depends on the parameter record {sc}; on the datums
    {d0x,d0y} of {xp} and {yp} associated by rung {g0}; on the datums
    {d1x,d1y} associated by rung {g1}; on the two datums {e0x,e0y}
    just after {d0x,d0y}; and on the two datums {e1x,e1y} just before
    {d1x,d1y}.
    
    The score {S} is the sum of a /data-related term/ {Sd}, that takes
    into account the similarity between those four datums pairs; and
    of a /move-related term/ {Sm} that takes into account the index
    increments {kx,ky} defined by the step along {xp} and {yp}. */

double dm_score_step_data_term
  ( dm_score_args_t *sc,
    msm_rung_t g0, 
    msm_rung_t g1, 
    dm_seq_t *xp, 
    dm_seq_t *yp
  );
  /* Computes the data-related part {Sd} of the score of a step
    {g0-->g1} between sequences {*xp} and {*yp}, according to the
    scoring criterion {sc}. The result measures the similarity between
    the datums paired along the step. */

double dm_score_step_move_term(dm_score_args_t *sc, msm_rung_t g, msm_rung_t h);
  /* Computes the move-related part {Sm} of the score of a step
    {g0-->g1} between sequences {*xp} and {*yp}, according to the
    scoring criterion {sc}. The result measures the shortness and
    even-ness of the step's motion along the two sequences. */

/* LOW_LEVEL SCORING FUNCTIONS

  The procedures in this section compute partial terms of 
  {dm_score_rung} and {dm_score_step}. */

double dm_score_rung_from_datums(double eql, double dif, dm_datum_t *dx, dm_datum_t *dy);
  /* Computes the score of a rung {g} that pairs a datum {dx} with a datum{dy},
    according to the scoring weights {eql} (match) and {dif} (mismatch). */

double dm_score_step_from_datums
  ( double eql, 
    double dif,
    dm_datum_t *d0x, 
    dm_datum_t *d0y, 
    dm_datum_t *d1x,
    dm_datum_t *d1y
  );
  /* Computes the data-related score for a step {g0-->g1},
    with scoring weights {eql} and {dif}.
    
    The parameters {d0x,d0y} and {d1x,d1y} are the datums paired by
    rungs {g0} and {g1}, respectively, in the two sequences.
    
    The step score is defined as the mean of
    {dm_score_rung_from_datums(eql,dif,d0x,d0y)} and
    {dm_score_rung_from_datums(eql,dif,d1x,d1y)}. Note that this
    function returns only the data-related part of the step's score,
    without the move-related part.  */

double dm_score_step_from_increments(double brk, double skp, double unp, int kx, int ky);
  /* Computes the move-related score of a step {g0-->g1} that advances
    {kx = msm_rung_X(g1) - msm_rung_X(g0)} units in the X-coordinate,
    and {ky = msm_rung_Y(g1) - msm_rung_Y(g0)} units in the Y-coordinate,
    according to penalty weights {brk}, {skp} and {unp}. */

/* COMMAND LINE ARGUMENT PARSING */

dm_score_args_t dm_score_args_parse(argparser_t *pp);
  /* Parses a scoring criterion {sc} from the command line argument
    parser {pp}, in the format described by {dm_score_args_HELP} and
    {dm_score_args_INFO} below. */

#define dm_score_args_HELP \
  "data {EQL_SCORE} {DIF_SCORE} move {BRK_SCORE} {SKP_SCORE} {UNP_SCORE}"

#define dm_score_args_INFO \
  dm_score_rung_data_args_INFO "\n" \
  "\n" \
  dm_score_step_data_args_INFO "\n" \
  "\n" \
  dm_score_step_move_args_INFO
 
#define dm_score_rung_data_args_INFO \
  "!!! Revise for subsampling !!!\n" \
  "The parameters {EQL_SCORE} and {DIF_SCORE} specify" \
  " the /data-related part/ of the score, which depends on the" \
  " similarity of the datums being paired.\n" \
  "\n" \
  "  The data-related score of a rung is a function {sd(dx,dy)} of the" \
  " two datums {dx,dy} paired by that rung.  If the samples" \
  " of {dx} and {dy} are all {+1} or {-1})," \
  " as in raw (unfiltered) sequences, the score" \
  " is {EQL_SCORE} if {dx == dy}, and {DIF_SCORE} if" \
  " {dx != dy}.   The parameter {EQL_SCORE} must be positive," \
  " and {DIF_SCORE} must be negative.\n" \
  "\n" \
  "  The function {sd(dx,dy)} is extended to" \
  " the fractional datums that occur in filtered" \
  " sequences by an appropriate quadratic estimation formula."

#define dm_score_step_data_args_INFO \
  "From the rung scoring function {sd} one derives another scoring" \
  " function {Sd} which is defined for any step between" \
  " rungs {g0} and {g1}.\n" \
  "\n" \
  "  If the step is  \"perfect\", i.e. it advances by exactly" \
  " 1 in each sequence, the data-related step score {Sd} depends only on the two" \
  " datums {d0x,d0y} paired by {g0} and the two datums {d1x,d1y}" \
  " paired by {g1}.  In that case, {Sd} is is simply the average" \
  " of {Sd(dx,dy)} as {dx} interpolates linearly from {d0x}" \
  " to {d1x}, and {dy} interpolates linearly from {d0y} to {d1y}.\n" \
  "\n" \
  "  If the step is \"imperfect\", the score {Sd} depends also on the two" \
  " datums datums {e0x,e0y} that immediately follow {d0x,d0y}, and" \
  " the two datums {e1x,e1y} that immediately precede {d1x,d1y}.  All" \
  " other datums that lie between the two rungs are ignored.\n" \
  "\n" \
  "   The data-related score for a list of rungs is simply the sum" \
  " of {Sd} over all steps of that list."

#define dm_score_step_move_args_INFO \
  "The parameters {BRK_SCORE}, {SKP_SCORE}, and" \
  " {UNP_SCORE} specify the /move-related part/ of" \
  " the score, which depends on the spacing between successive" \
  "  rungs along the two sequences.\n" \
  "\n" \
  "  The move-related score is defined for any 1-increasing" \
  " step as a function {Sm(kx,ky)} of the index increments" \
  " {kx} and {ky} implied by the step along" \
  " the two sequences.\n" \
  "\n" \
  "  If the step is \"perfect\"," \
  " namely if {kx == ky == 1}, the score {Sm(kx,ky)} is zero.  Otherwise, the" \
  " score is\n" \
  "\n" \
  "     {BRK_SCORE + min(kx-1,ky-1)*SKP_SCORE + abs(kx-ky)*UNP_SCORE}\n" \
  "\n" \
  "  Here {BRK_SCORE} is the basic (negative) penalty for any \"imperfect\" step;" \
  " {SKP_SCORE} is the additional (negative) penalty for skipping over" \
  " one datum in both sequences; and {UNP_SCORE} the additional (negative)" \
  " penalty for skipping one datum in one sequence only.\n" \
  "\n" \
  "   The move-related score for a list of rungs is simply the sum" \
  " of {Sm} over all steps of that list."
  
#endif
