#define PROG_NAME "dm_test_010_filter"
#define PROG_DESC "test of DNA signal filtering routines"
#define PROG_VERS "1.0"

/* Last edited on 2009-02-23 22:21:12 by stolfi */

#define dm_test_010_filter_C_COPYRIGHT \
  "Copyright  2006  by the State University of Campinas (UNICAMP)"

#define PROG_HELP \
  PROG_NAME " \\\n" \
  "  -maxLevel {MAX_LEVEL} \\\n" \
  "  [ -circular ] \\\n" \
  "  -initFilter " wt_table_args_HELP " \\\n" \
  "  -incrFilter " wt_table_args_HELP " \\\n" \
  "  {SEQ_FILE} \\\n" \
  "  {OUT_NAME}"

#define PROG_INFO \
  "NAME\n" \
  "  " PROG_NAME " - " PROG_DESC "\n" \
  "\n" \
  "SYNOPSIS\n" \
  "  " PROG_HELP "\n" \
  "\n" \
  "DESCRIPTION\n" \
  "  This program reads a" \
  " DNA/RNA sequence from {SEQ_FILE}, converts it" \
  " to numeric format, and outputs" \
  " filtered and subsampled versions of that" \
  " sequence.  The input file should be in \".seq\" format" \
  " (see the INPUT FILE FORMAT section below).\n" \
  "\n" \
  "INPUT FILE FORMAT\n" \
  dm_nucleic_file_format_INFO "\n" \
  "\n" \
  "OUTPUT FILES\n" \
  "  All output files will have names starting with {OUT_NAME}.\n" \
  "\n" \
  "OPTIONS\n" \
  "  -circular\n" \
  "    If present, this keyword specifies that the" \
  " sequence in {SEQ_FILE} is circular (like a bacterial" \
  " chromosome or plasmid).\n" \
  "\n" \
  "  -maxLevel {MAX_LEVEL}\n" \
  "    This mandatory argument specifies the" \
  " maximum level of filtering to be" \
  " applied.  Level 0 is the unfiltered" \
  " sequence, and each subsequent level is sampled at half" \
  " the frequency of the previous one.\n" \
  "\n" \
  "  -initFilter " wt_table_args_HELP " \n" \
  "  -incrFilter " wt_table_args_HELP " \n" \
  "    These mandatory arguments specify the" \
  " weights of the filter to be used at the" \
  " first filtering step and at subsequent" \
  " filtering steps.  " wt_table_args_norm_sum_INFO "\n" \
  "\n" \
  argparser_help_info_HELP_INFO "\n" \
  "SEE ALSO\n" \
  "  dm_match(1)\n" \
  "\n" \
  "AUTHOR\n" \
  "  This program was created on 16/dec/2006 by J. Stolfi.\n" \
  "WARRANTY\n" \
  argparser_help_info_NO_WARRANTY "\n" \
  "\n" \
  "RIGHTS\n" \
  "  " dm_test_010_filter_C_COPYRIGHT ".\n" \
  "\n" \
  argparser_help_info_STANDARD_RIGHTS  

#include <dm_basic.h>
#include <dm_seq.h>
#include <dm_filter.h>
#include <dm_nucleic.h>
#include <dm_test_tools.h>

#include <msm_test_tools.h>
#include <msm_multi.h>

#include <wt_table.h>
#include <vec.h>
#include <affirm.h>
#include <argparser.h>

#include <math.h>
#include <assert.h>
#include <string.h>
#include <stdlib.h>

typedef struct options_t 
  { char *seqName;     /* Name of file with DNA/RNA sequence (minus ".bas" ext). */
    bool_t circular;   /* TRUE iff sequence is circular. */
    int maxLevel;      /* Maximum filtering level. */
    double_vec_t w0;   /* Filter weights for first stage (level 0 to level 1). */
    double_vec_t w1;   /* Filter weights for subsequent stages. */
    char *outName;     /* Output file name prefix (minus extensions). */
  } options_t;
  
int main(int argc, char**argv);

options_t *dm_get_options(int argc, char**argv);
  /* Parses the command line options, packs 
    them into a {options_t} record. */

int main(int argc, char**argv)
  { 
    options_t *o = dm_get_options(argc, argv);

    fprintf(stderr, "reading input sequence ...\n");
    dm_seq_t s = dm_seq_read_from_nucleic_file(0, "s", o->seqName, o->circular);
    fprintf(stderr, "sequence has %d sample datums\n", dm_seq_num_datums(&s));
    
    fprintf(stderr, "getting the weight tables ...\n");
    char *wname0 = wt_table_make_descr(o->w0.ne, o->w0.e, "%6.4f");
    char *wname1 = wt_table_make_descr(o->w1.ne, o->w1.e, "%6.4f");

    fprintf(stderr, "trying to filter up to level %d ...\n", o->maxLevel);
    dm_seq_t r[o->maxLevel + 1];
    dm_filter_seq_multi(&s, o->maxLevel, &(o->w0), wname0, &(o->w1), wname1, r, 1);
    
    /* Figure out max useful level {maxLevel}: */
    int maxLevel = o->maxLevel;
    while ((maxLevel >= 0) && (dm_seq_num_datums(&(r[maxLevel])) == 0)) { maxLevel--; }
    fprintf(stderr, "maximum non-empty level is %d\n", maxLevel);
    
    fprintf(stderr, "writing and plotting the filtered sequences ...\n");
    dm_test_tools_seq_multi_write_and_plot_named(r, maxLevel, "test seq", o->outName, "-ot");

    /* ...and we are done: */
    fprintf(stderr, "done.\n");
    return 0;
  }

options_t* dm_get_options(int argc, char**argv)
  { options_t *o = (options_t *)notnull(malloc(sizeof(options_t)), "no mem");
    
    argparser_t *pp = argparser_new(stderr, argc, argv);
    argparser_set_help(pp, PROG_HELP);
    argparser_set_info(pp, PROG_INFO);
    argparser_process_help_info_options(pp);
    
    o->circular = argparser_keyword_present(pp, "-circular");

    argparser_get_keyword(pp, "-initFilter");
    o->w0 = wt_table_args_parse(pp, TRUE);

    argparser_get_keyword(pp, "-incrFilter");
    o->w1 = wt_table_args_parse(pp, TRUE);

    argparser_get_keyword(pp, "-maxLevel");
    o->maxLevel = argparser_get_next_int(pp, 0, 20);

    argparser_skip_parsed(pp);
    
    o->seqName = argparser_get_next(pp);
    
    o->outName = argparser_get_next(pp);

    argparser_finish(pp);
    
    return o;
  }
