#define PROG_NAME "dm_test_012_fourier"
#define PROG_DESC "Fourier analysis of DNA signal filtering routines"
#define PROG_VERS "1.0"

/* Last edited on 2009-02-23 22:22:39 by stolfi */

#define dm_test_012_fourier_C_COPYRIGHT \
  "Copyright  2006  by the State University of Campinas (UNICAMP)"

#define PROG_HELP \
  PROG_NAME " \\\n" \
  "  -maxLevel {MAX_LEVEL} \\\n" \
  "  -numPoints {N_POINTS} \\\n" \
  "  -initFilter " wt_table_args_HELP " \\\n" \
  "  -incrFilter " wt_table_args_HELP " \\\n" \
  "  {OUT_NAME}"

#define PROG_INFO \
  "NAME\n" \
  "  " PROG_NAME " - " PROG_DESC "\n" \
  "\n" \
  "SYNOPSIS\n" \
  "  " PROG_HELP "\n" \
  "\n" \
  "DESCRIPTION\n" \
  "  This program analyzes the spectrum of numeric DNA" \
  " signals filtered at various scales with given" \
  " filter coeffcients." \
  " .\n" \
  "\n" \
  "INPUT FILE FORMAT\n" \
  dm_nucleic_file_format_INFO "\n" \
  "\n" \
  "OUTPUT FILES\n" \
  "  All output files will have names starting with {OUT_NAME}.\n" \
  "\n" \
  "OPTIONS\n" \
  "  -maxLevel {MAX_LEVEL}\n" \
  "    This mandatory argument specifies the" \
  " maximum level of filtering to be" \
  " applied.  Level 0 is the unfiltered" \
  " sequence, and each subsequent level is sampled at half" \
  " the frequency of the previous one.\n" \
  "\n" \
  "  -numPoints {N_POINTS}\n" \
  "    This mandatory argument specifies the" \
  " number of data points (nucleotide bases)" \
  " to use at the lowest (unfiltered) scale.\n" \
  "\n" \
  "  -initFilter " wt_table_args_HELP " \n" \
  "  -incrFilter " wt_table_args_HELP " \n" \
  "    These mandatory arguments specify the" \
  " weights of the filter to be used at the" \
  " first filtering step and at subsequent" \
  " filtering steps.  " wt_table_args_norm_sum_INFO "\n" \
  "\n" \
  argparser_help_info_HELP_INFO "\n" \
  "SEE ALSO\n" \
  "  dm_match(1)\n" \
  "\n" \
  "AUTHOR\n" \
  "  This program was created on 27/dec/2006 by J. Stolfi.\n" \
  "WARRANTY\n" \
  argparser_help_info_NO_WARRANTY "\n" \
  "\n" \
  "RIGHTS\n" \
  "  " dm_test_012_fourier_C_COPYRIGHT ".\n" \
  "\n" \
  argparser_help_info_STANDARD_RIGHTS  

#include <dm_basic.h>
#include <dm_seq.h>
#include <dm_nucleic.h>
#include <dm_test_tools.h>

#include <msm_ps_tools.h>
#include <msm_multi.h>

#include <wt_table.h>
#include <vec.h>
#include <affirm.h>
#include <argparser.h>

#include <fftw3.h>

#include <math.h>
#include <assert.h>
#include <string.h>
#include <stdlib.h>

typedef struct options_t 
  { int maxLevel;       /* Maximum filtering level. */
    int numPoints;      /* Number of bases in test sequence. */
    double_vec_t w0;    /* Filter weights for first stage (level 0 to level 1). */
    double_vec_t w1;    /* Filter weights for subsequent stages. */
    char *outName;      /* Output file name prefix (minus extensions). */
  } options_t;
  
int main(int argc, char**argv);

void dm_compute_random_dna_spectrum(int N, double dnaP[]);
  /* Stores in {dnaP[0..N/2]} the theoretical power spectrum of a
    random DNA sequence. Assumes that the bases are uniformly
    distributed and independent. Also assumes that they are converted
    to numeric 3-vectors {(1,1,1)} as described under
    {dm_datum_decoded_from_nucleic_char}. The number {N} must be even.
    
    The spectrum is the sum of the spectra of all three channels */

void dm_compute_filter_spectrum(int nw, double w[ ], int N, double wP[]);
  /* Stores in {wP[0..N/2]} the power spectrum of the sequence {w[0..nw-1]},
    zero-extended to {N} samples.  Requires {nw <= N}. */

void dm_dump_dft(int N, char *xN, fftw_complex *xP, char *yN, fftw_complex *yP);
  /* Writes the complex vectors {xP[0..N-1]} and {yP[0..N-1]}
    (typically, a complex discrete signal and its DFT), side by side.
    The parameters {xN,yN} should be the names of the three vectors.
    If either vector is NULL, it is omitted from the printout. */

void dm_dump_spectra(int fMax, char *xN, double *xP, char *yN, double *yP, char *zN, double *zP);
  /* Writes the power spectra {xP[0..fMax]}, {yP[0..fMax]},
    {zP[0..fMax]} side by side. The parameters {xN,yN,zN} should be
    the names of the three vectors. If either vector is NULL, it is
    omitted from the printout. */

void dm_plot_spectra
  ( double *xP,
    double *yP,
    double *zP,
    int fMax,
    int fSep,
    double vMax,
    char *outName,
    char *tag,
    int level
  );
  /* Plots up to three power spectra on the same graph, using the same 
    scales.  
    
    The X axis runs from 0 to {fMax}, the Y axis from 0 to {vMax} or
    the maximum power in the data, whichever is greater.
    
    The power spectra are assumed to be {xP[0..fMax]},
    {yP[0..fMax]}, {zP[0..fMax]}.  If any of {xP,yP,zP} is NULL, the 
    corresponding graph is omitted.  A vertical line is drawn at 
    frequency [fSep}.  The plot is written to file
    "{outName}{tag}-{level}.eps". */

options_t *dm_get_options(int argc, char**argv);
  /* Parses the command line options, packs 
    them into a {options_t} record. */

#define dm_debug_fft FALSE

int main(int argc, char**argv)
  { 
    options_t *o = dm_get_options(argc, argv);

    /* Number of points at base level: */
    int N = o->numPoints;
    
    fprintf(stderr, "reading input sequence ...\n");
    /* dm_seq_t s = dm_seq_read_from_nucleic_file(0, "s", o->seqName, o->circular); */

    /* Maximum frequency in power spectra: */
    int fMax = N/2;
    
    /* Nominal maximum power in random DNA spectrum: */
    double dnaMax = 6.0*dm_NUCLEIC_RAW_SCALE*dm_NUCLEIC_RAW_SCALE;
    
    /* Nominal maximum power in filter spectra: */
    double wMax = 1.0;
    
    fprintf(stderr, "computing the mean power spectrum of a random DNA sequence...\n");
    double dnaP[fMax+1];
    dm_compute_random_dna_spectrum(N, dnaP);
    dm_plot_spectra(dnaP, NULL, NULL, fMax, 0, dnaMax, o->outName, "-db", 0);
    
    fprintf(stderr, "computing the power spectra of the two filters ...\n");
    wt_table_print(stderr, "initial", o->w0.ne, o->w0.e);
    double w0P[fMax+1];
    dm_compute_filter_spectrum(o->w0.ne, o->w0.e, N, w0P);
    dm_plot_spectra(w0P, NULL, NULL, fMax, 0, wMax, o->outName, "-f0", 0);

    wt_table_print(stderr, "incremental", o->w1.ne, o->w1.e);
    double w1P[fMax+1];
    dm_compute_filter_spectrum(o->w1.ne, o->w1.e, N, w1P);
    dm_plot_spectra(w1P, NULL, NULL, fMax, 0, wMax, o->outName, "-f1", 0);
    
    if (dm_debug_fft) 
      { fprintf(stderr, "normalized power spectra of input data:\n");
        dm_dump_spectra(fMax, "dnaP", dnaP, "w0P ", w0P, "w1P ", w1P);
      }

    fprintf(stderr, "running Fourier analysis up to level %d ...\n", o->maxLevel);
    
    /* !!! Should account for quantization error too !!! */
    
    /* Valid (non-aliased) spectrum of filtered DNA signal: */
    int f;
    double *vP = (double *)notnull(malloc((fMax+1)*sizeof(double)), "no mem");
    for (f = 0; f < fMax; f++) { vP[f] = dnaP[f] * w0P[f]; }
    /* Invalid (aliased) spectrum of filtered DNA signal: */
    double *aP = (double *)notnull(malloc((fMax+1)*sizeof(double)), "no mem"); 
    for (f = 0; f < fMax; f++) { aP[f] = 0; }
    
    /* Current filtering level: */
    int level = 0;
    /* Current max freq in {vP,aP}: */
    int gMax = fMax;
    while (TRUE)
      { /* Sampling step relative to base level: */
        int step = (1 << level);
        assert(gMax = N/step/2);
        
        /* Debugging: */
        if (dm_debug_fft)
          { fprintf(stderr, "power spectra of preserved and aliased components:\n");
            dm_dump_spectra(gMax, "vP", vP, "aP", aP, NULL, NULL);
          }

        /* Plot spectra of preserved and aliased signals: */
        dm_plot_spectra(vP, aP, NULL, gMax, 0, dnaMax, o->outName, "-ss", level);

        /* Compute the current S,N powers and N/S ratio: */
        double vPt = 0; /* Total power in non-aliased freqencies. */
        double aPt = 0; /* Power in aliased frequencies. */
        int g;
        for (g = 0; g < gMax; g++) { vPt += vP[g]; aPt += aP[g]; }
        fprintf
          ( stderr,
            "total power S+N = %8.5f  valid S = %8.5f  aliased N = %8.5f  N/(S+N) = %8.6f\n",
            vPt+aPt, vPt, aPt, aPt/(vPt+aPt)
          );

        /* Can we do another stage of filtering? */
        if (level >= o->maxLevel) { break; }
        if (gMax % 2 != 0) { break; }
        if (o->w1.ne > N/step) { break; }

        /* Compute power spectrum of incremental filter: */
        double wincP[gMax+1];
        dm_compute_filter_spectrum(o->w1.ne, o->w1.e, N/step, wincP);
        dm_plot_spectra(wincP, NULL, NULL, gMax, 0, wMax, o->outName, "-ft", level+1);
        if (dm_debug_fft)
          { fprintf(stderr, "power spectrum of incremental filter:\n");
            dm_dump_spectra(fMax, "wincP", wincP, NULL, NULL, NULL, NULL);
          }
        
        /* Compute max frequency {hMax} after downsampling: */
        int hMax = gMax/2;
        assert(hMax == N/step/4);
        
        /* Apply incremental filter to {vP,aP}: */
        for (g = 0; g <= gMax; g++) { vP[g] *= wincP[g]; aP[g] *= wincP[g]; }
        dm_plot_spectra(vP, aP, NULL, gMax, hMax, wMax, o->outName, "-sf", level+1);

        /* Allocate new storage for valid and aliased spectra after downsampling: */
        double *vfP =  (double *)notnull(malloc((hMax+1)*sizeof(double)), "no mem");
        double *afP =  (double *)notnull(malloc((hMax+1)*sizeof(double)), "no mem");
        
        /* Simulate downsampling of {vP,aP} to {vfP,afP}: */
        int h;
        for (h = 0; h <= hMax; h++) { vfP[h] = afP[h] = 0; }
        for (g = 0; g <= gMax; g++)
          { /* Compute the frequency {h} that {g} becomes when 2-sampled: */
            h = 2*g;
            if (h > gMax) h = h - 2*gMax;
            if (h < 0) { h = -h; }
            assert(h % 2 == 0);
            h /= 2;
            assert((h >= 0) && (h <= hMax));
            /* Add power {vP[g]}  to {vfP[g]} or {afP[g]} as appropriate: */
            if (g == h) 
              { /* Freq {g} gets preserved: */      vfP[h] += vP[g]; }
            else
              { /* Freq {g} gets aliased to {h}: */ afP[h] += vP[g]; }
            /* Add aliased power {aP[g]} to {afP[h]}: */
            afP[h] += aP[g];
          }
          
        /* Prepare for next iteration: */
        free(vP); vP = vfP;
        free(aP); aP = afP;
        level++;
      }

    /* ...and we are done: */
    fprintf(stderr, "done.\n");
    return 0;
  }

void dm_dump_spectra(int fMax, char *xN, double *xP, char *yN, double *yP, char *zN, double *zP)
  { int f;
    for (f = 0; f <= fMax; f++)
      { if (xP != NULL) { fprintf(stderr, " %s[%4d] = %8.6f", xN, f, xP[f]); }
        if (yP != NULL) { fprintf(stderr, " %s[%4d] = %8.6f", yN, f, yP[f]); }
        if (zP != NULL) { fprintf(stderr, " %s[%4d] = %8.6f", zN, f, zP[f]); }
        fprintf(stderr, "\n");
      }
  }

void dm_dump_dft(int N, char *xN, fftw_complex *xP, char *yN, fftw_complex *yP)
  { int i;
    for (i = 0; i < N; i++)
      { if (xP != NULL)
          { fprintf(stderr, "  %s[%4d] = ( %+9.6f %+9.6f )", xN, i, xP[i][0], xP[i][1]); }
        if (yP != NULL) 
          { fprintf(stderr, "  %s[%4d] = ( %+9.6f %+9.6f )", yN, i, yP[i][0], yP[i][1]); }
        fprintf(stderr, "\n"); 
      }
  }

void dm_compute_random_dna_spectrum(int N, double dnaP[])
  { /* Get and check some parameters: */
    demand(N % 2 == 0, "N must be even");
    int fMax = N/2; /* Maximum frequency n power spectrum. */
    int nc = dm_CHANNELS; /* Number of channels in datums. */
    assert(nc == 3); /* The analysis below is specific for {nc=3}. */
    double uP = 1.0; /* Expected power per channel and per freq {0..N-1} */
    /* In each channel, the samples are {-1} or {+1} with equal prob. */
    /* The expected value of the mean squared is {1/N}: */
    dnaP[0] = nc*uP;
    /* For frequency {fMax}, the expected power is {1/N} too: */
    dnaP[fMax] = nc*uP;
    /* For other freqs {1..fMax-1}, the expected power is {2/N}  too: */
    int f;
    for (f = 1; f < fMax; f++) { dnaP[f] = 6*uP; }
  }

void dm_compute_filter_spectrum(int nw, double w[ ], int N, double wP[])
  { /* Get and check some parameters: */
    demand(N % 2 == 0, "N must be even");
    int fMax = N/2; /* Maximum frequency in power spectrum. */
    demand(nw <= N, "nw must not exceed N");
    
    /* Allocate vectors {in,ot} for {fftw3.h} routines: */
    fftw_complex *in = fftw_malloc(sizeof(fftw_complex)*N); /* Input vector. */
    fftw_complex *ot = fftw_malloc(sizeof(fftw_complex)*N); /* Output vector. */

    /* Precompute parameters for {fftw3.h} routines: */
    fftw_plan plan = fftw_plan_dft_1d(N, in, ot, FFTW_FORWARD, FFTW_ESTIMATE);

    /* Store weights {w} in input vector, as complex numbers: */
    int i;
    for (i = 0; i < N; i++)
      { in[i][0] = (i < nw ? w[i] : 0); 
        in[i][1] = 0; 
      }

    /* Compute Fourier transform: */
    fftw_execute(plan);
      
    if (dm_debug_fft)
      { fprintf(stderr, "input and output of Fouier transform:\n");
        dm_dump_dft(N, "in", in, "ot", ot);
      }

    /* Compute the power spectrum: */
    double norm = 1/((double)N); /* Power normalization factor. */
    int f;
    for (f = 0; f <= fMax; f++)
      { double P = ot[f][0]*ot[f][0] + ot[f][1]*ot[f][1];
        if ((f > 0) && (f < fMax))
          { int g = N-f;
            P += ot[g][0]*ot[g][0] + ot[g][1]*ot[g][1];
          }
        wP[f] = norm*P;
      }
      
    if (dm_debug_fft)
      { fprintf(stderr, "raw power spectrum:\n");
        for (f = 0; f <= fMax; f++)
          { fprintf(stderr, " wP[%4d] = %8.6f\n", f, wP[f]); }
      }

    /* Normalize power spectrum so that the max elem is 1: */
    double vmax = -INF;
    for (f = 0; f <= fMax; f++) { if (wP[f] > vmax) { vmax = wP[f]; } }
    for (f = 0; f <= fMax; f++) { wP[f] /= vmax; }
      
    /* Cleanup: */
    fftw_destroy_plan(plan);
    fftw_free(in); 
    fftw_free(ot);
  }

void dm_plot_spectra
  ( double *xP,
    double *yP,
    double *zP,
    int fMax,
    int fSep,
    double vMax,
    char *outName,
    char *tag,
    int level
  )
  { /* Colors for curves (all with brightness = 0.30): */
    double R[3] = { 1.00, 0.00, 0.00 };
    double G[3] = { 0.00, 0.50, 0.33 };
    double B[3] = { 0.00, 0.00, 1.00 };
    
    /* Figure dimensions in mm: */
    double mrgL = 15.0; /* Width of Y scale and ticks. */
    double mrgB =  4.0; /* Height of X scale and ticks. */
    double mrgR =  3.0; /* Possible right overflow of X scale. */
    double mrgT =  2.0; /* Possible top overflow of Y scale. */
    double hTotSize = 180.0 + mrgL + mrgR; /* Tot width of plottable area (incl scales). */
    double vTotSize =  30.0 + mrgB + mrgT; /* Tot height of plottable area (incl scales). */
    double mrgX = 1.0; /* Extra margin around plottable area. */

    /* Create plotter object {dp} and get its Postscript stream {ps}: */
    char *levTag = NULL;
    asprintf(&levTag, "%s-%02d", tag, level);
    msm_ps_tools_t *dp = 
      msm_ps_tools_new
        ( NULL, outName, levTag, hTotSize, vTotSize, mrgX, dm_test_tools_FONT_SIZE );
    PSStream *ps = msm_ps_tools_get_ps_stream(dp);
    
    /* Choose nominal X range {[xMin_xMax]}: */
    double xSkosh = 0.01*fMax;
    double xMin = 0 - xSkosh;
    double xMax = fMax + xSkosh;
    
    /* Enlarge {vmax} if needed to enclose all data points: */
    int f;
    for (f = 0; f <= fMax; f++)
      { if ((xP != NULL) && (xP[f] > vMax)) { vMax = xP[f]; } 
        if ((yP != NULL) && (yP[f] > vMax)) { vMax = yP[f]; } 
        if ((zP != NULL) && (zP[f] > vMax)) { vMax = zP[f]; } 
      }

    /* Compute nominal Y range {[yMin_yMax]}: */
    double ySkosh = 0.02*vMax;
    double yMax = vMax + ySkosh;
    double yMin = 0 - ySkosh;
    
    /* Set client reference window, with some skosh: */
    msm_ps_tools_set_client_ref_window(dp, xMin, xMax, yMin, yMax);
    
    /* Shrink device window to leave space for tick labels: */
    msm_ps_tools_shrink_device_ref_window(dp, mrgL, mrgR, mrgB, mrgT);
    
    /* Decide whether to plot dots at individual samples: */
    double hStep = msm_ps_tools_map_x(dp, 1) - msm_ps_tools_map_x(dp, 0);
    bool_t show_dots = (hStep >= 1.5);  /* If spaced at least 1.5 mm */

    /* Draw axes, ticks, labels, etc: */
    char *font = "Times-Roman";
    double fontSz = 12.0; /* In Postscript points. */
    double tickSz = 1.0;  /* In mm. */
    pswr_set_label_font(ps, font, fontSz);
    /* msm_ps_tools_draw_ref_axis(dp, HOR); */
    /* msm_ps_tools_draw_ref_axis(dp, VER); */
    msm_ps_tools_draw_scale(dp, HOR, yMin, tickSz,0.0, 5.0,1.0, "%.0f",  1.2, 12.0,0.0);
    msm_ps_tools_draw_scale(dp, VER, xMin, tickSz,0.0, 3.0,0.0, "%+.2f", 1.2,  4.0,0.0);
    
    if ((fSep > 0) && (fSep < fMax))
      { /* Draw a vertical line at {fSep}: */
        pswr_set_pen(ps, 1.000,0.750,0.000, 0.25, 0.0, 0.0);
        msm_ps_tools_draw_segment(dp, (double)fSep, yMin, (double)fSep, yMax);
      }
    
    /* Draw polylines: */
    if (xP != NULL) 
      { pswr_set_pen(ps, R[0],G[0],B[0], 0.25, 0.0, 0.0);
        msm_ps_tools_draw_y_polyline(dp, 0, fMax, xP, fMax+1);
      }
    if (yP != NULL) 
      { pswr_set_pen(ps, R[1],G[1],B[1], 0.25, 0.0, 0.0);
        msm_ps_tools_draw_y_polyline(dp, 0, fMax, yP, fMax+1);
      }
    if (zP != NULL) 
      { pswr_set_pen(ps, R[2],G[2],B[2], 0.25, 0.0, 0.0);
        msm_ps_tools_draw_y_polyline(dp, 0, fMax, zP, fMax+1);
      }
    
    if (show_dots)
      { /* Draw dots: */
        if (xP != NULL) 
          { pswr_set_fill_color(ps, R[0],G[0],B[0]);
            msm_ps_tools_draw_y_dots(dp, 0, fMax, xP, fMax+1, 0.5, TRUE, FALSE);
          }
        if (yP != NULL) 
          { pswr_set_fill_color(ps, R[1],G[1],B[1]);
            msm_ps_tools_draw_y_dots(dp, 0, fMax, yP, fMax+1, 0.5, TRUE, FALSE);
          }
        if (zP != NULL) 
          { pswr_set_fill_color(ps, R[2],G[2],B[2]);
            msm_ps_tools_draw_y_dots(dp, 0, fMax, zP, fMax+1, 0.5, TRUE, FALSE);
          }
      }

    msm_ps_tools_draw_ref_frame(dp);
    msm_ps_tools_close(dp);
    free(levTag);
  }

options_t* dm_get_options(int argc, char**argv)
  { options_t *o = (options_t *)notnull(malloc(sizeof(options_t)), "no mem");
    
    argparser_t *pp = argparser_new(stderr, argc, argv);
    argparser_set_help(pp, PROG_HELP);
    argparser_set_info(pp, PROG_INFO);
    argparser_process_help_info_options(pp);
    
    argparser_get_keyword(pp, "-maxLevel");
    o->maxLevel = argparser_get_next_int(pp, 0, 20);

    argparser_get_keyword(pp, "-numPoints");
    o->numPoints = argparser_get_next_int(pp, 0, (1 << 18));

    argparser_get_keyword(pp, "-initFilter");
    o->w0 = wt_table_args_parse(pp, TRUE);

    argparser_get_keyword(pp, "-incrFilter");
    o->w1 = wt_table_args_parse(pp, TRUE);

    argparser_skip_parsed(pp);
    
    /* o->seqName = argparser_get_next(pp); */
    
    o->outName = argparser_get_next(pp);

    argparser_finish(pp);
    
    return o;
  }
