#define PROG_NAME "dm_test_120_distance"
#define PROG_DESC "test of DNA signal distance functions"
#define PROG_VERS "1.0"

/* Last edited on 2009-02-23 22:23:08 by stolfi */

#define dm_test_distance_C_COPYRIGHT \
  "Copyright  2006  by the State University of Campinas (UNICAMP)"

#define PROG_HELP \
  PROG_NAME " \\\n" \
  "  -trials {N_TRIALS} \\\n" \
  "  -filterWeights " wt_table_args_HELP " \\\n" \
  "  -diffSqrWeights " wt_table_args_HELP " \\\n" \
  "  [ -verbose ] \\\n" \
  "  {OUT_NAME}"

#define PROG_INFO \
  "NAME\n" \
  "  " PROG_NAME " - " PROG_DESC "\n" \
  "\n" \
  "SYNOPSIS\n" \
  "  " PROG_HELP "\n" \
  "\n" \
  "DESCRIPTION\n" \
  "  This program tests distance estimators" \
  " under various filter and distance weights.\n" \
  "\n" \
  "  Suppose we have two sequences {A[0..N-1],B[0..N-1]}" \
  " which were generated by filtering two DNA strings {a[0..n-1],b[0..n-1]}" \
  " with smoothing filter weights {ws[0..m-1]}.  Let {wd[0..m-1]} be another" \
  " set of weights.  We wish to estimate the local average of the squared" \
  " difference {D2[i] = SUM{wd[j]*|a[i+j]-b[i+j]|^2 : j}} from the two" \
  " smoothed samples {A[i],B[i]}.\n" \
  "\n" \
  "OUTPUT FILES\n" \
  "  All output files will have names starting with {OUT_NAME}.\n" \
  "\n" \
  "OPTIONS\n" \
  "  -trials {N_TRIALS}\n" \
  "    This mandatory argument specifies the" \
  " number of test cases to generate.\n" \
  "\n" \
  "  -filterWeights " wt_table_args_HELP "\n" \
  "  -diffSqrWeights " wt_table_args_HELP "\n" \
  "    These arguments specify the" \
  " weights to be used to compute the smoothed DNA signals" \
  " and the smoothed squared difference signal," \
  " respectively.  The two lists must have the" \
  " same number of elements.  The" \
  " \"-filterWeights\" option is" \
  " mandatory; if \"-diffSqrWeights\" is omitted," \
  " the squared diff weights default to the" \
  " same as the signal" \
  " weights.  " wt_table_args_norm_sum_INFO ".\n" \
  "\n" \
  "  -verbose\n" \
  "    If this option is present, prints detailed debuging info.\n" \
  "\n" \
  argparser_help_info_HELP_INFO "\n" \
  "SEE ALSO\n" \
  "  dm_match(1)\n" \
  "\n" \
  "AUTHOR\n" \
  "  This program was created on 21/dec/2006 by J. Stolfi.\n" \
  "WARRANTY\n" \
  argparser_help_info_NO_WARRANTY "\n" \
  "\n" \
  "RIGHTS\n" \
  "  " dm_test_distance_C_COPYRIGHT ".\n" \
  "\n" \
  argparser_help_info_STANDARD_RIGHTS  

#include <math.h>
#include <assert.h>
#include <string.h>
#include <stdlib.h>

#include <dm_basic.h>
#include <dm_seq.h>
#include <dm_nucleic.h>

#include <msm_test_tools.h>

#include <jsrandom.h>
#include <bool.h>
#include <affirm.h>
#include <argparser.h>
#include <gauss_elim.h>
#include <lsq.h>
#include <wt_table.h>

typedef struct options_t 
  { double_vec_t ws;   /* Filter weights for DNA signals. */
    double_vec_t wd;   /* Filter weights for squared difference signal. */
    int trials;        /* Number of test cases to generate. */
    bool_t verbose;    /* TRUE to print debugging info. */
    /* Output parameters: */
    char *outName;     /* Output file name prefix (minus extensions). */
  } options_t;
  
int main(int argc, char**argv);

int dm_td_num_basis_funs(int nw);
  /* Number of linearly independent 
    functions in the least-squares fitting basis. */

void dm_td_fit_dist_formulas(double_vec_t *wsv, double_vec_t *wdv, int nt, char *outName, bool_t verbose);
  /* Estimates the optimum distance formula for the weight vector {w},
    using {nt} test cases. */

void dm_td_throw_case
  ( double ws[],
    double wd[],
    int nw,
    double X[],
    double Y[],
    double Z[],
    bool_t verbose
  );
  /* Generates two random nucleotide strings with {nw} letters.
    Computes their numeric equivalents {x[0..nw-1]}
    and {y[0..nw-1]}, where each {x[i]} or {y[i]} is a 3-vector 
    with the convention described under {dm_datum_decoded_from_nucleic_char}.
    Then computes the filter outputs {X = SUM{ ws[i]*x[i] : i in 0..nw-1}}
    and similarly for Y.  Also computes {Z[i] = |x[i]-y[i]|^2} for 
    {i} in {0..nw-1}, and {Z[nw] = SUM{wd[i]*Z[i] : i in 0..nw-1}}.
    
    If {verbose} is true, writes the test case to {stderr}. */

void dm_td_apply_filter(int nw, double w[], int *x, double X[]);
  /* Applies a filter with weights {w[0..nw-1]} to the columns of 
    the integer array {x}.  Assumes that {x} has {nw} rows and {nc} columns,
    where {nc=dm_CHANNELS}. Returns the result in the float vector 
    {X[0..nc-1]}. */

void dm_normalize_doubles(double v[], int nv);
  /* Scales {v[0..nv-1]} so that they add to 1. */

double dm_doubles_normsq(double v[], int nv);
  /* Returns the sum of the squares of the values {v[0..nv-1]}. */

double dm_td_lsq_basis(int ib, double X[], double Y[]);
  /* Evaluates basis function number {ib} on {X} and {Y},
    for {ib} in {0..dm_BASIS_FUNCTIONS-1}. */

char *dm_td_lsq_basis_txt(int ib);
  /* Retursn a description of basis function number {ib},
    for {ib} in {0..dm_BASIS_FUNCTIONS-1}. */

options_t *msm_get_options(int argc, char **argv);
  /* Parses the command line options, packs 
    them into a {options_t} record. */

int main(int argc, char**argv)
  { 
    options_t *o = msm_get_options(argc, argv);
    
    dm_td_fit_dist_formulas(&(o->ws), &(o->wd), o->trials, o->outName, o->verbose);

    return 0;
  }
  
void dm_td_fit_dist_formulas(double_vec_t *wsv, double_vec_t *wdv, int nt, char *outName, bool_t verbose)
  { 
    fprintf(stderr, "----------------------------------------------------------------------\n");
    fprintf(stderr, "Testing distances...\n");
    fprintf(stderr, "...\n");

    demand(wsv->ne == wdv->ne, "weight lists must have the same length");
    
    /* Get problem dimensions: */
    int nw = wsv->ne;                    /* Number of entries in signal weight table. */
    int nc = dm_CHANNELS;                 /* Number of channels in datum. */
    int nz = nw+1;                        /* Number of functions to approximate. */
    int nb = dm_td_num_basis_funs(nw);    /* Number of relevant basis functions. */
    
    /* Print parameters and weight vector: */
    fprintf(stderr, "nt = %d\n", nt);
    fprintf(stderr, "nw = %d\n", nw);

    /* Normalize DNA weights to unit sum: */
    double ws[nw];
    int iw; for (iw = 0; iw < nw; iw++) { ws[iw] = wsv->e[iw]; }
    dm_normalize_doubles(ws, nw);
    fprintf(stderr, "normalized signal weights:\n");
    fprintf(stderr, "ws = ");
    msm_debug_double_vec(ws, nw, "%10.8f"); 
    fprintf(stderr, " var = %12.10f", dm_doubles_normsq(ws, nw));
    fprintf(stderr, "\n");
    
    /* Normalize squared diff weights to unit sum: */
    double wd[nw];
    for (iw = 0; iw < nw; iw++) { wd[iw] = wdv->e[iw]; }
    dm_normalize_doubles(wd, nw);
    fprintf(stderr, "normalized squared diff weights:\n");
    fprintf(stderr, "wd = ");
    msm_debug_double_vec(wd, nw, "%10.8f"); 
    fprintf(stderr, " var = %12.10f", dm_doubles_normsq(wd, nw));
    fprintf(stderr, "\n");
    
    auto void gen_case(int it, int nv, double v[], int nf, double f[]);
      /* Case generator. */
      
    void gen_case(int it, int nv, double v[], int nf, double f[])
      { assert(nv == nb);
        assert(nf == nz);
        double X[nc], Y[nc];
        bool_t verbgen = verbose & (it < 2); /* Debug the test case generator? */
        dm_td_throw_case(ws, wd, nw, X, Y, f, verbgen); 
        int ib;
        for (ib = 0; ib < nb; ib++) { v[ib] = dm_td_lsq_basis(ib, X, Y); }
      }
    
    double U[nb*nz];
    int rank = lsq_fit(nb, nz, nt, gen_case, U, verbose);
    demand(rank == nb, "could not solve the least-squares system");
    
    /* Print the fitted formula: */
    int iz;
    for (iz = 0; iz < nz; iz++)
      { fprintf(stderr, "best-fit estimator for Z[%d] = ", iz);
        if (iz < nw)
          { fprintf(stderr, "|x[%d] - y[%d]|^2", iz, iz); }
        else if (iz == nw)
          { fprintf(stderr, "S2(x,y)"); }
        else
          { assert(FALSE); }
        fprintf(stderr, ":\n");
        int ib;
        for (ib = 0; ib < nb; ib++)
          { char *op = (ib == nb-1 ? "" : " +");
            fprintf(stderr, "  %+12.8f*(%s)%s\n", U[ib*nz + iz], dm_td_lsq_basis_txt(ib), op);
          }
        fprintf(stderr, "\n");
      }
    fprintf(stderr, "----------------------------------------------------------------------\n");
  }
    
double dm_td_lsq_basis(int ib, double X[], double Y[])
  { double X0 = X[0], X1 = X[1], X2 = X[2];
    double Y0 = Y[0], Y1 = Y[1], Y2 = Y[2];
    switch(ib)
      { case 0: 
          return 1;
        case 1: 
          return X0*Y0 + X1*Y1 + X2*Y2;
        case 2: 
          return X0*X0 + X1*X1 + X2*X2 + Y0*Y0 + Y1*Y1 + Y2*Y2;
        case 3: 
          return X0 + X1 + X2 + Y0 + Y1 + Y2;
        case 4: 
          return X0*X1 + X1*X2 + X2*X0 + Y0*Y1 + Y1*Y2 + Y2*Y0;
        case 5: 
          return X0*Y1 + X1*Y2 + X2*Y0 + X1*Y0 + X2*Y1 + X0*Y2;
        default: 
          affirm(FALSE, "bad ib"); return 0;
      }
  }
    
char *dm_td_lsq_basis_txt(int ib)
  { switch(ib)
      { case 0: return "1";
        case 1: return "<X|Y>";
        case 2: return "(|X|^2 + |Y|^2)";
        case 3: return "(X0 + X1 + X2 + Y0 + Y1 + Y2)";
        case 4: return "(X0*X1 + X1*X2 + X2*X0 + Y0*Y1 + Y1*Y2 + Y2*Y0)";
        case 5: return "(X0*Y1 + X1*Y2 + X2*Y0 + X1*Y0 + X2*Y1 + X0*y2)";
        default: affirm(FALSE, "bad ib"); return "BAD";
      }
  }

int dm_td_num_basis_funs(int nw)
 { 
   if (nw == 1) 
     { return 2; }
   else if (nw == 2)
     { return 6; /* ??? Not sure... */ }
   else
     { return 6; }
 }

void dm_td_throw_case
  ( double ws[],
    double wd[],
    int nw,
    double X[],
    double Y[],
    double Z[],
    bool_t verbose
  )
  {      
    /* Get problem dimensions: */
    int nc = dm_CHANNELS; /* Number of channels in datum. */
      
    /* DNA alphabet: */
    char *atcg = "ATCG"; /* The four nucleotides. */
    
    /* The two DNA letter sequences: */
    char xc[nw + 1]; xc[nw] = 0; /* First nucleotide string (with final 0). */
    char yc[nw + 1]; yc[nw] = 0; /* Second nucleotide string (with final 0). */

    /* The numeric coordinates of {xc[0..nw-1],yc[0..nw-1]}: */
    int x[nw*nc];  /* Element {x[i*nc+c]} is channel {c} of numerized {xc[i]}. */ 
    int y[nw*nc];  /* Element {y[i*nc+c]} is channel {c} of numerized {yc[i]}. */

    /* Generate two random nucleotide strings {x,y}: */
    randchars(xc, nw, atcg, 4);
    randchars(yc, nw, atcg, 4);
    if (verbose) 
      { fprintf(stderr, "  x = %s", xc);
        fprintf(stderr, "  y = %s", yc);
        fprintf(stderr, "\n"); 
      }
    /* Convert them to numeric vectors: */
    int iw;
    for (iw = 0; iw < nw; iw++) 
      { dm_datum_decoded_from_nucleic_char(xc[iw], &(x[iw*nc]));
        dm_datum_decoded_from_nucleic_char(yc[iw], &(y[iw*nc]));
        if (verbose)
          { fprintf(stderr, "  x[%02d] = %c = ", iw, xc[iw]); 
            msm_debug_int_vec(&(x[iw*nc]), nc, "%+d");
            fprintf(stderr, "  y[%02d] = %c = ", iw, yc[iw]); 
            msm_debug_int_vec(&(y[iw*nc]), nc, "%+d");
            fprintf(stderr, "\n"); 
          }
      }
    /* Compute the decoded non-quantized filtered datums {X,Y}: */
    int c;
    dm_td_apply_filter(nw, ws, x, X);
    dm_td_apply_filter(nw, ws, y, Y);
    if (verbose)
      { fprintf(stderr, "  X = "); 
        msm_debug_double_vec(X, nc, "%+10.6f");
        fprintf(stderr, "  Y = "); 
        msm_debug_double_vec(X, nc, "%+10.6f");
        fprintf(stderr, "\n"); 
      }
    
    /* Compute target quantities {Z[i]}: */
    double sumd2 = 0;
    for (iw = 0; iw < nw; iw++)
      { double sumd2i = 0;
        for (c = 0; c < nc; c++)
          { int d = x[iw*nc+c] - y[iw*nc+c]; sumd2i += d*d; }
        Z[iw] = sumd2i;
        sumd2 += sumd2i*wd[iw];
      }
    Z[nw] = sumd2;
  }

void dm_normalize_doubles(double v[], int nv)
  { double sum = 0;
    int i;
    for (i = 0; i < nv; i++) { sum += v[i]; }
    if (sum == 0) { return; }
    for (i = 0; i < nv; i++) { v[i] /= sum; }
  }

double dm_doubles_normsq(double v[], int nv)
  { int i;
    double sum2 = 0;
    for (i = 0; i < nv; i++) { double d = v[i]; sum2 += d*d; }
    return sum2;
  }

void dm_td_apply_filter(int nw, double w[], int *x, double X[])
  { int nc = dm_CHANNELS;
    int c;
    for (c = 0; c < nc; c++)
      { double sumwx = 0;
        int iw;
        for (iw = 0; iw < nw; iw++)
          { sumwx += w[iw]*((double)x[iw*nc+c]); }
        X[c] = sumwx;
      }
  }

options_t *msm_get_options(int argc, char**argv)
  { 
    options_t *o = (options_t *)notnull(malloc(sizeof(options_t)), "no mem");
    
    argparser_t *pp = argparser_new(stderr, argc, argv);
    argparser_set_help(pp, PROG_HELP);
    argparser_set_info(pp, PROG_INFO);
    argparser_process_help_info_options(pp);
    
    argparser_get_keyword(pp, "-filterWeights");
    o->ws = wt_table_args_parse(pp, TRUE);
    
    if (argparser_keyword_present(pp, "-diffSqrWeights"))
      { o->wd = wt_table_args_parse(pp, TRUE); 
        if (o->wd.ne != o->ws.ne) 
          { argparser_error(pp, "weight lists must have the same length"); }
      }
    else
      { o->wd = o->ws; }
    
    argparser_get_keyword(pp, "-trials");
    int ntmin = dm_td_num_basis_funs(o->wd.ne); /* Lsq needs at least this many data points */
    o->trials = argparser_get_next_double(pp, ntmin, INT_MAX);
    
    o->verbose = argparser_keyword_present(pp, "-verbose");
    
    argparser_skip_parsed(pp);
    
    o->outName = argparser_get_next(pp);

    argparser_finish(pp);
    
    return o;
  }
