#define PROG_NAME "dm_test_scoring"
#define PROG_DESC "test of rung and step scoring functions"
#define PROG_VERS "1.0"

/* Last edited on 2008-01-30 17:09:08 by hcgl */

#define dm_test_scoring_C_COPYRIGHT \
  "Copyright  2006  by the State University of Campinas (UNICAMP)"

#define PROG_HELP \
  PROG_NAME " \\\n" \
  "  -scores " dm_score_args_HELP " \\\n" \
  "  [ -symmetry ] \\\n" \
  "  {OUT_NAME}"

#define PROG_INFO \
  "NAME\n" \
  "  " PROG_NAME " - " PROG_DESC "\n" \
  "\n" \
  "SYNOPSIS\n" \
  "  " PROG_HELP "\n" \
  "\n" \
  "DESCRIPTION\n" \
  "  This program prints the numerical score for various" \
  " rungs and steps between typical datums.\n" \
  "OUTPUT FILES\n" \
  "  All output files will have names starting with {OUT_NAME}.\n" \
  "\n" \
  "OPTIONS\n" \
  "  -scores " dm_score_args_HELP "\n" \
  "    This mandatory argument specifies the" \
  " scoring criterion to be used when evaluating" \
  " and comparing candidates.  " dm_score_args_INFO "\n" \
  "\n" \
  "  -symmetry\n" \
  "    If this optional argument is present, the" \
  " program will check whether the scoring functions" \
  " are symmetric under X-Y exchange.  If the" \
  " argument is omitted, the program" \
  " assumes the functions are symmetric, and skips the" \
  " testing of rungs and steps that differ only by X-Y" \
  " exchange.\n" \
  "\n" \
  argparser_help_info_HELP_INFO "\n" \
  "SEE ALSO\n" \
  "  dm_match(1)\n" \
  "\n" \
  "AUTHOR\n" \
  "  This program was created on 21/dec/2006 by J. Stolfi.\n" \
  "WARRANTY\n" \
  argparser_help_info_NO_WARRANTY "\n" \
  "\n" \
  "RIGHTS\n" \
  "  " dm_test_scoring_C_COPYRIGHT ".\n" \
  "\n" \
  argparser_help_info_STANDARD_RIGHTS  

#include <dm_basic.h>
#include <dm_seq.h>
#include <dm_nucleic.h>
#include <dm_score.h>

#include <msm_rung.h>

#include <bool.h>
#include <affirm.h>
#include <argparser.h>

#include <math.h>
#include <assert.h>
#include <string.h>
#include <stdlib.h>

typedef struct options_t 
  { dm_score_args_t sc;   /* Scoring criterion. */
    bool_t symmetry;      /* If TRUE also checks X-Y symmetry. */
    /* Output parameters: */
    char *outName;     /* Output file name prefix (minus extensions). */
  } options_t;
  
int main(int argc, char**argv);

dm_datum_t dm_datum_rescale(double alpha, dm_datum_t d) ;
  /* Multiplies all channels of datum {d} by {alpha}. Note that if {d}
   is not in {-1,0,+1}, the result will almost always include some
   quantization error. */

void dm_test_all_rung_scores(double mag, bool_t sym, dm_score_args_t *sc);
  /* Tests the single-rung scoring functions for many
    nucleotide pairs, with sample values scaled by {mag}.
    If {sym} is TRUE also checks whether the scores are
    symmetric under X-Y exchange. */

double dm_test_rung_score(double mag, dm_score_args_t *sc, char cx, char cy);
  /* Tests the single-rung scoring functions for the 
    nucleotide pair {cx,cy}, with sample values scaled by {mag}.
    Returns the step's score. */

void dm_test_all_quadruplet_scores(double mag, bool_t sym, dm_score_args_t *sc);
  /* Tests the step scoring functions for many 
    pairs of nucleotide pairs, with sample values scaled by
    {mag}.  If {sym} is TRUE also checks whether the scores are
    symmetric under X-Y exchange. */

double dm_test_quadruplet_score
  ( double mag, 
    dm_score_args_t *sc, 
    char ax, 
    char ay, 
    char bx, 
    char by
  );
  /* Tests the step scoring functions for various ways of pairing four
    bases {ux ax bx vx} in the first sequence with four bases {uy
    ay by vy} in the second sequence, assuming that (1) {ux} is
    paired with {uy} and {vx} is paired with {vy}, and (2) the
    steps are strictly increasing along both sequences.
    
    There are six possible pairings with these constraints, which can be 
    represented as "||||", "||:|", "|:||", "|'|.|", "|.|'|", "|::|".
    
    The scores for all these alternatives includes the data-score of
    pairing {ux} with {uy} and {vx} with {vy}; therefore we can
    leave this term out, and we do not need to know those four bases.
    The rest of the score consists of the data-score for the other
    paired datums, and the move-score for the internal terms. Returns
    a weighted average of these six scores. */

options_t *msm_get_options(int argc, char**argv);
  /* Parses the command line options, packs 
    them into a {options_t} record. */

int main(int argc, char**argv)
  { 
    options_t *o = msm_get_options(argc, argv);
    
    dm_test_all_rung_scores(1.0, o->symmetry, &(o->sc));
    dm_test_all_rung_scores(0.5, o->symmetry, &(o->sc));

    dm_test_all_quadruplet_scores(1.0, o->symmetry, &(o->sc));
    dm_test_all_quadruplet_scores(0.5, o->symmetry, &(o->sc));

    return 0;
  }
  
dm_datum_t dm_datum_rescale(double alpha, dm_datum_t d) 
  { int c;
    for (c = 0; c < dm_CHANNELS; c++) 
      { d.c[c] = dm_sample_encode(alpha*dm_sample_decode(d.c[c]));}
    return d;
  }

void dm_test_all_rung_scores(double mag, bool_t sym, dm_score_args_t *sc)
  { char *b = "ATCG"; /* The four nucleotides. */
    
    fprintf(stderr, "----------------------------------------------------------------------\n");
    fprintf(stderr, "Testing rung scores (mag = %8.6f) ...\n", mag);
    fprintf(stderr, "...\n");
    int x, y;
    /* Enumerate all rungs {(x,y)} : */
    for (x = 0; x < 4; x++)
      { for (y = x; y < 4; y++)
          { double scXY = dm_test_rung_score(mag, sc, b[x], b[y]);
            if (sym && (x != y)) 
              { /* Scoring should be symmetric, but let's check: */
                double scYX = dm_test_rung_score(mag, sc, b[y], b[x]);
                affirm(scXY == scYX, "rung scoring is not X-Y symmetric");
              }
            fprintf(stderr, "\n");
          }
      }
    fprintf(stderr, "----------------------------------------------------------------------\n");
  }

double dm_test_rung_score(double mag, dm_score_args_t *sc, char cx, char cy)
  { dm_datum_t dx = dm_datum_rescale(mag, dm_datum_encoded_from_nucleic_char(cx));
    dm_datum_t dy = dm_datum_rescale(mag, dm_datum_encoded_from_nucleic_char(cy));
    double diffsq = dm_datum_diffsq(&dx, &dy);
    double score = dm_score_rung_from_datums(sc->eql, sc->dif, &dx, &dy);
    fprintf(stderr, "%c = ", cx);
    dm_datum_decoded_write(stderr, &dx, "( ", " ", " )");
    fprintf(stderr, "  ");
    fprintf(stderr, "%c = ", cy);
    dm_datum_decoded_write(stderr, &dy, "( ", " ", " )");
    fprintf(stderr, "  ");
    fprintf(stderr, " d2 = %8.6f ", diffsq);
    fprintf(stderr, " sc = %+9.6f", score),  
    fprintf(stderr, "\n");
    return score;
  }

void dm_test_all_quadruplet_scores(double mag, bool_t sym, dm_score_args_t *sc)
  { char *b = "ATCG"; /* The four nucleotides. */
    
    fprintf(stderr, "----------------------------------------------------------------------\n");
    fprintf(stderr, "Testing step scores (mag = %8.6f) ...\n", mag);
    fprintf(stderr, "...\n");
    int iax, iay, ibx, iby;
    /* Enumerate all quadruplets {iax,ibx,iay,iby} where {(iax,ibx)<=(iay,iby)}: */
    for (iax = 0; iax < 4; iax++)
      { for (iay = iax; iay < 4; iay++)
          { for (ibx = 0; ibx < 4; ibx++)
              { for (iby = (iax == iay ? ibx : 0); iby < 4; iby++)
                  { double scXY = dm_test_quadruplet_score(mag, sc, b[iax], b[iay], b[ibx], b[iby]);
                    if (sym && ((iax != iay) || (ibx != iby)))
                      { /* Scoring should be symmetric, but let's check: */
                        double scYX = dm_test_quadruplet_score(mag, sc, b[iay], b[iax], b[iby], b[ibx]);
                        affirm(scXY == scYX, "step scoring is not X-Y symmetric");
                      }
                    fprintf(stderr, "\n");
                  }
              }
          }
      }
    fprintf(stderr, "----------------------------------------------------------------------\n");
  }

#define MAX_TERMS 10
  /* Upper bound to number of terms in a quadruplet score. */

double dm_test_quadruplet_score
  ( double mag, 
    dm_score_args_t *sc, 
    char ax, 
    char ay, 
    char bx, 
    char by
  )
  { /* Get the four datums {dax,day,dbx,dby}: */
    dm_datum_t dax = dm_datum_rescale(mag, dm_datum_encoded_from_nucleic_char(ax));
    dm_datum_t day = dm_datum_rescale(mag, dm_datum_encoded_from_nucleic_char(ay));
    dm_datum_t dbx = dm_datum_rescale(mag, dm_datum_encoded_from_nucleic_char(bx));
    dm_datum_t dby = dm_datum_rescale(mag, dm_datum_encoded_from_nucleic_char(by));
    fprintf(stderr, "%c %c = ", ax, bx);
    dm_datum_decoded_write(stderr, &dax, "( ", " ", " )");
    dm_datum_decoded_write(stderr, &dbx, "( ", " ", " )");
    fprintf(stderr, "\n");
    fprintf(stderr, "%c %c = ", ay, by);
    dm_datum_decoded_write(stderr, &day, "( ", " ", " )");
    dm_datum_decoded_write(stderr, &dby, "( ", " ", " )");
    fprintf(stderr, "\n");
    
    /* Mean difference squared along step: */
    double diffsq = dm_datum_step_diffsq(&dax, &day, &dbx, &dby);
    fprintf(stderr, "  step d2 = %8.6f\n", diffsq);
    
    /* Score terms for debugging: */
    double score; /* The total score for one possible pairing. */
    int nterms;  /* Number of terms that contributed to the score. */
    double term[MAX_TERMS]; /* The terms are {term[0..nterms-1]}. */
    char type[MAX_TERMS];   /* The type of {term[i]} is {type[i]}. */
    
    double comb_score = 0; /* A combination of the scores of all six possible pairings. */
    
    auto void begin_score(char *pr);
      /* Starts computing the score for the pairing with the
        schema {pr}. */
    
    auto void step_score(int sx, int sy);
      /* Computes the move-score of a step that advances {sx} in the 
        first sequence and {sy} in the second sequence,and appends 
        it to {term[0..nterms-1]}, with type 'M'. */
    
    auto void data_score(dm_datum_t dx, dm_datum_t dy);
      /* Computes the data-score of a rung that pairs {cx} in the 
        first sequence and {cy} in the second sequence, and appends 
        it to {term[0..nterms-1]}, with type 'D'. */
    
    auto void end_score(double wt);
      /* Finishes the computation of a score, prints it, and all 
        its terms. Also combines {score} into {comb_score} with 
        weight {wt}. */
    
    begin_score("||||");
    step_score(1,1);
    data_score(dax,day);
    step_score(1,1);
    data_score(dbx,dby);
    step_score(1,1);
    end_score(10000);
    
    begin_score("||:|");
    step_score(1,1);
    data_score(dax,day);
    step_score(2,2);
    end_score(1000);
    
    begin_score("|:||");
    step_score(2,2);
    data_score(dbx,dby);
    step_score(1,1);
    end_score(100);
    
    begin_score("|'|.|");
    step_score(2,1);
    data_score(dbx,day);
    step_score(1,2);
    end_score(10); /* Same weight as "|.|'|" to pass the X-Y symmetry test. */
    
    begin_score("|.|'|");
    step_score(1,2);
    data_score(dax,dby);
    step_score(2,1);
    end_score(10); /* Same weight as "|'|.|" to pass the X-Y symmetry test. */
    
    begin_score("|::|");
    step_score(3,3);
    end_score(1);
    
    /* Returns the average score: */
    fprintf(stderr, "  combined score = %+9.2f\n", comb_score);
    fprintf(stderr, "\n");
    return comb_score;
    
    auto void begin_score(char *pr)
      { fprintf(stderr, "  score for %-6s", pr);
        score = 0; nterms = 0;
      }
    
    auto void step_score(int sx, int sy)
      { double t = dm_score_step_from_increments(sc->brk, sc->skp, sc->unp, sx,sy);
        score += t;
        assert(nterms < MAX_TERMS);
        term[nterms] = t;
        type[nterms] = 'M';
        nterms++;
      }
    
    auto void data_score(dm_datum_t dx, dm_datum_t dy)
      { double t = dm_score_rung_from_datums(sc->eql, sc->dif, &dx, &dy);
        score += t;
        assert(nterms < MAX_TERMS);
        term[nterms] = t;
        type[nterms] = 'D';
        nterms++;
      }
    
   auto void end_score(double wt)
      { fprintf(stderr, " = %+9.6f = ", score);
        int k;
        for (k = 0; k < nterms; k++)
          { fprintf(stderr, " %+6.3f(%c)", term[k], type[k]); }
        fprintf(stderr, "\n");
        /* Accumulate in {comb_scores}: */
        comb_score += wt*score;
      }
   }
    
options_t *msm_get_options(int argc, char**argv)
  { 
    options_t *o = (options_t *)notnull(malloc(sizeof(options_t)), "no mem");
    
    argparser_t *pp = argparser_new(stderr, argc, argv);
    argparser_set_help(pp, PROG_HELP);
    argparser_set_info(pp, PROG_INFO);
    argparser_process_help_info_options(pp);
    
    o->symmetry = argparser_keyword_present(pp, "-symmetry");
    
    argparser_get_keyword(pp, "-scores");
    o->sc = dm_score_args_parse(pp);
    
    argparser_skip_parsed(pp);
    
    o->outName = argparser_get_next(pp);

    argparser_finish(pp);
    
    return o;
  }
